"""
File Writer for scenarios to commonroad xml-format
"""
import datetime
import enum
import pathlib
import os
from typing import Union, List, Set
import numpy as np
import decimal
import warnings

from commonroad import SCENARIO_VERSION
from lxml import etree, objectify

from commonroad.common.util import Interval
from commonroad.geometry.shape import Rectangle, Circle, Polygon, ShapeGroup
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem
from commonroad.prediction.prediction import SetBasedPrediction, TrajectoryPrediction
from commonroad.scenario.intersection import Intersection
from commonroad.scenario.lanelet import Lanelet, LineMarking, StopLine, LaneletType
from commonroad.scenario.obstacle import ObstacleRole, ObstacleType, DynamicObstacle, StaticObstacle, Obstacle, \
    Occupancy, Shape, SignalState, EnvironmentObstacle, PhantomObstacle
from commonroad.scenario.scenario import Scenario, Tag, Location, GeoTransformation, Environment, Weather, TimeOfDay, \
    Underground
from commonroad.scenario.traffic_sign import TrafficSign, TrafficLight, TrafficLightCycleElement, TrafficLightDirection
from commonroad.scenario.trajectory import Trajectory, State

__author__ = "Stefanie Manzinger, Moritz Klischat, Sebastian Maierhofer"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "2021.2"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"


class DecimalPrecision:
    decimals = 4


precision = DecimalPrecision


def float_to_str(f):
    """
    Convert the given float to a string,
    without resorting to scientific notation
    """
    fstring = str(f)
    if "e" in fstring:
        return format(f, ".{}f".format(precision.decimals))
    f_list = fstring.split(".")
    if len(f_list) > 1:
        return f_list[0] + "." + f_list[1][:precision.decimals]
    else:
        return f_list[0]


def create_exact_node_float(value: Union[int, float]) -> etree.Element:
    """
    creates element node for exact value
    :param value: exact value
    :return: node for exact value
    """
    node = etree.Element('exact')
    node.text = float_to_str(np.float64(value))
    return node


def create_exact_node_int(value: Union[int]) -> etree.Element:
    """
    creates element node for exact value
    :param value: exact value
    :return: node for exact value
    """
    assert np.issubdtype(type(value), np.integer), (
        '<util/create_exact_node_int> expected type int for value but'
        ' got %s' % (type(value))
    )
    node = etree.Element('exact')
    node.text = str(value)
    return node


def create_interval_node_float(interval: Interval) -> List[etree.Element]:
    """
    creates ElementTree.Element for an interval
    :param interval:
    :return: list of Element nodes with start and end of interval
    """
    node_lo = etree.Element('intervalStart')
    node_lo.text = float_to_str(np.float64(interval.start))
    node_hi = etree.Element('intervalEnd')
    node_hi.text = float_to_str(np.float64(interval.end))
    return [node_lo, node_hi]


def create_interval_node_int(interval: Interval) -> List[etree.Element]:
    """
    creates ElementTree.Element for an interval
    :param interval:
    :return: list of Element nodes with start and end of interval
    """
    assert np.issubdtype(type(interval.start), np.integer), (
        '<util/create_exact_node_int> expected type int for value but got %s'
        % (type(interval.start))
    )
    assert np.issubdtype(type(interval.end), np.integer), (
        '<util/create_exact_node_int> expected type int for value but got %s'
        % (type(interval.start))
    )

    node_lo = etree.Element('intervalStart')
    node_lo.text = str(interval.start)
    node_hi = etree.Element('intervalEnd')
    node_hi.text = str(interval.end)
    return [node_lo, node_hi]


class OverwriteExistingFile(enum.Enum):
    """
    Specifies whether an existing file will be overwritten or skipped
    """

    ASK_USER_INPUT = 0
    ALWAYS = 1
    SKIP = 2


class CommonRoadFileWriter:
    def __init__(
        self,
        scenario: Scenario,
        planning_problem_set: PlanningProblemSet,
        author: str = None,
        affiliation: str = None,
        source: str = None,
        tags: Set[Tag] = None,
        location: Location = None,
        decimal_precision: int = 4,
    ):
        """
        Initialize the FileWriter with a scenario and tags for the xml-header

        :param scenario: scenario that should be written later
        :param planning_problem_set: corresponding planning problem to the scenario
        :param author: author's name
        :param affiliation: affiliation of the author
        :param source: source of dataset (d.h. database, handcrafted, etc.)
        :param tags: list of keywords describing the scenario (e.g. road type(one-lane road, multilane),
                required maneuver etc., see commonroad.in.tum.de for full list))
        :param decimal_precision: number of decimal places used when writing float values
        """
        assert not (author is None and scenario.author is None)
        assert not (affiliation is None and scenario.affiliation is None)
        assert not (source is None and scenario.source is None)
        assert not (tags is None and scenario.tags is None)

        self.scenario: Scenario = scenario
        self.planning_problem_set: PlanningProblemSet = planning_problem_set
        self._root_node = etree.Element('commonRoad')
        self.author = author if author is not None else scenario.author
        self.affiliation = affiliation if affiliation is not None else scenario.affiliation
        self.source = source if source is not None else scenario.source
        self.location = location if location is not None else scenario.location
        self.tags = tags if tags is not None else scenario.tags

        # set decimal precision
        precision.decimals = decimal_precision

    @property
    def root_node(self):
        return self._root_node

    @root_node.setter
    def root_node(self, root_node):
        warnings.warn(
            '<CommonRoadFileWriter/root_node> root_node of CommonRoadFileWriter is immutable'
        )

    @property
    def author(self):
        return self._author

    @author.setter
    def author(self, author):
        assert isinstance(
            author, str
        ), '<CommonRoadFileWriter/author> author must be a string, but has type {}'.format(
            type(author)
        )
        self._author = author

    @property
    def affiliation(self):
        return self._affiliation

    @affiliation.setter
    def affiliation(self, affiliation):
        assert isinstance(
            affiliation, str
        ), '<CommonRoadFileWriter/affiliation> affiliation must be a string, but has type {}'.format(
            type(affiliation)
        )
        self._affiliation = affiliation

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        assert isinstance(
            source, str
        ), '<CommonRoadFileWriter/source> source must be a string, but has type {}'.format(
            type(source)
        )
        self._source = source

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        for tag in tags:
            assert isinstance(tag, Tag), '<CommonRoadFileWriter/tags> tag must ' \
                                         'be a enum of type Tag, but has type {}'.format(type(tag))
        self._tags = tags

    def _write_header(self):
        self._root_node.set('timeStepSize', str(self.scenario.dt))
        self._root_node.set('commonRoadVersion', SCENARIO_VERSION)
        self._root_node.set('author', self.author)
        self._root_node.set('affiliation', self.affiliation)
        self._root_node.set('source', self.source)

        try:
            if self.scenario.scenario_id:
                self._root_node.set('benchmarkID', str(self.scenario.scenario_id))
        except:
            self._root_node.set('benchmarkID', '-1')
            print('Warning: No scenario_id set.')

        self._root_node.set('date', datetime.datetime.today().strftime('%Y-%m-%d'))

    def _add_all_objects_from_scenario(self):
        if self.location is not None:
            self._root_node.append(LocationXMLNode.create_node(self.location))
        else:
            self._root_node.append(LocationXMLNode.create_node(Location()))
        self._root_node.append(TagXMLNode.create_node(self.tags))
        for l in self.scenario.lanelet_network.lanelets:
            self._root_node.append(LaneletXMLNode.create_node(l))
        for sign in self.scenario.lanelet_network.traffic_signs:
            self._root_node.append(TrafficSignXMLNode.create_node(sign))
        for light in self.scenario.lanelet_network.traffic_lights:
            self._root_node.append(TrafficLightXMLNode.create_node(light))
        for intersection in self.scenario.lanelet_network.intersections:
            self._root_node.append(IntersectionXMLNode.create_node(intersection))
        for o in self.scenario.obstacles:
            self._root_node.append(ObstacleXMLNode.create_node(o))

    def _add_all_planning_problems_from_planning_problem_set(self):
        for (
            planning_problem
        ) in self.planning_problem_set.planning_problem_dict.values():
            self._root_node.append(PlanningProblemXMLNode.create_node(planning_problem))

    def _dump(self):
        rough_string = etree.tostring(
            self._root_node, pretty_print=True, encoding='UTF-8'
        )
        rough_string = rough_string
        return rough_string

    def write_to_file(
        self,
        filename: Union[str, None] = None,
        overwrite_existing_file: OverwriteExistingFile = OverwriteExistingFile.ASK_USER_INPUT,
        check_validity: bool = False
    ):
        """
        Write a scenario including planning-problem. If file already exists, it will be overwritten of skipped

        :param filename: filename of the xml output file. If 'None', the Benchmark ID is taken
        :param overwrite_existing_file: Specify whether an already existing file should be overwritten or skipped
        :param check_validity: check xml file against .xsd definition
        :return:
        """
        if filename is None:
            filename = str(self.scenario.scenario_id)

        if pathlib.Path(filename).is_file():
            if overwrite_existing_file is OverwriteExistingFile.ASK_USER_INPUT:
                overwrite = input(
                    'File {} already exists, replace old file (or else skip)? (y/n)'.format(
                        filename
                    )
                )
            elif overwrite_existing_file is OverwriteExistingFile.SKIP:
                overwrite = 'n'
            else:
                overwrite = 'y'

            if overwrite == 'n':
                print('Writing of file {} skipped'.format(filename))
                return
            else:
                print('Replace file {}'.format(filename))

        self._write_header()
        self._add_all_objects_from_scenario()
        self._add_all_planning_problems_from_planning_problem_set()
        if check_validity:
            # validate xml format
            self.check_validity_of_commonroad_file(self._dump())

        tree = etree.ElementTree(self._root_node)
        tree.write(filename, pretty_print=True, xml_declaration=True, encoding="utf-8")

    def write_scenario_to_file(
        self,
        filename: Union[str, None] = None,
        overwrite_existing_file: OverwriteExistingFile = OverwriteExistingFile.ASK_USER_INPUT
    ):
        """
        Write a scenario without planning-problem. If file already exists, it will be overwritten of skipped.

        :param filename: filename of the xml output file. If 'None', the Benchmark ID is taken
        :param OverwriteExistingFile: Specify whether an already existing file should be overwritten or skipped
        :return: None
        """
        if filename is None:
            filename = str(self.scenario.scenario_id)

        if pathlib.Path(filename).is_file():
            if overwrite_existing_file is OverwriteExistingFile.ASK_USER_INPUT:
                overwrite = input(
                    'File {} already exists, replace old file (or else skip)? (y/n)'.format(
                        filename
                    )
                )
            elif overwrite_existing_file is OverwriteExistingFile.SKIP:
                overwrite = 'n'
            else:
                overwrite = 'y'

            if overwrite == 'n':
                print(
                    'Writing skipped for file, since it already exists {}'.format(
                        filename
                    )
                )
                return
            else:
                print('Replace file {}'.format(filename))

        self._write_header()
        self._add_all_objects_from_scenario()

        tree = etree.ElementTree(self._root_node)
        tree.write(filename, pretty_print=True, xml_declaration=True, encoding="utf-8")

    @staticmethod
    def check_validity_of_commonroad_file(commonroad_str: str):
        """Check the validity of a generated xml_string in terms of
        commonroad with an existing XSD schema.
        Throw an error if it is not valid.

        Args:
          commonroad_str: XML formatted string which should be checked.

        """
        with open(
            os.path.dirname(os.path.abspath(__file__)) + '/../xml_definition_files/XML_commonRoad_XSD.xsd',
            'rb',
        ) as schema_file:
            schema = etree.XMLSchema(etree.parse(schema_file))

        parser = objectify.makeparser(schema=schema, encoding='utf-8')

        try:
            etree.fromstring(commonroad_str, parser)
        except etree.XMLSyntaxError as error:
            raise Exception(
                'Could not produce valid CommonRoad file! Error: {}'.format(error.msg)
            )


class LocationXMLNode:
    @classmethod
    def create_node(cls, location: Location) -> etree.Element:
        """
        Create XML-Node for a location
        :param location: location object
        :return: node
        """
        location_node = etree.Element('location')
        geo_name_id_node = etree.Element("geoNameId")
        geo_name_id_node.text = str(location.geo_name_id)
        location_node.append(geo_name_id_node)
        gps_latitude_node = etree.Element("gpsLatitude")
        gps_latitude_node.text = str(location.gps_latitude)
        location_node.append(gps_latitude_node)
        gps_longitude_node = etree.Element("gpsLongitude")
        gps_longitude_node.text = str(location.gps_longitude)
        location_node.append(gps_longitude_node)
        if location.geo_transformation is not None:
            location_node.append(GeoTransformationXMLNode.create_node(location.geo_transformation))
        if location.environment is not None:
            location_node.append(EnvironmentXMLNode.create_node(location.environment))

        return location_node


class GeoTransformationXMLNode:
    @classmethod
    def create_node(cls, geo_transformation: GeoTransformation) -> etree.Element:
        """
        Create XML-Node for a location
        :param geo_transformation: GeoTransformation object
        :return: node
        """
        geotransform_node = etree.Element('geoTransformation')
        geo_reference_node = etree.Element("geoReference")
        geo_reference_node.text = geo_transformation.geo_reference
        geotransform_node.append(geo_reference_node)
        additional_transformation_node = etree.Element('additionalTransformation')
        x_translation_node = etree.Element("xTranslation")
        x_translation_node.text = str(geo_transformation.x_translation)
        additional_transformation_node.append(x_translation_node)
        y_translation_node = etree.Element("yTranslation")
        y_translation_node.text = str(geo_transformation.y_translation)
        additional_transformation_node.append(y_translation_node)
        z_rotation_node = etree.Element("zRotation")
        z_rotation_node.text = str(geo_transformation.z_rotation)
        additional_transformation_node.append(z_rotation_node)
        scaling_node = etree.Element("scaling")
        scaling_node.text = str(geo_transformation.scaling)
        additional_transformation_node.append(scaling_node)
        geotransform_node.append(additional_transformation_node)

        return geotransform_node


class EnvironmentXMLNode:
    @classmethod
    def create_node(cls, environment: Environment) -> etree.Element:
        """
        Create XML-Node for a environment
        :param environment: Environment object
        :return: node
        """
        environment_node = etree.Element('environment')
        if environment.time_of_day.value is not TimeOfDay.UNKNOWN:
            time_node = etree.Element('time')
            if environment.time.hours < 10:
                time_node.text = "0" + str(environment.time.hours) + ":" + str(environment.time.minutes) + ":00"
            else:
                time_node.text = str(environment.time.hours) + ":" + str(environment.time.minutes)
            environment_node.append(time_node)
            time_of_day_node = etree.Element('timeOfDay')
            time_of_day_node.text = environment.time_of_day.value
            environment_node.append(time_of_day_node)
        if environment.weather.value is not Weather.UNKNOWN:
            weather_node = etree.Element('weather')
            weather_node.text = environment.weather.value
            environment_node.append(weather_node)
        if environment.underground.value is not Underground.UNKNOWN:
            underground_node = etree.Element('underground')
            underground_node.text = environment.underground.value
            environment_node.append(underground_node)

        return environment_node


class TagXMLNode:
    @classmethod
    def create_node(cls, tags: Set[Tag]) -> etree.Element:
        """
        Create XML-Node for a tag element
        :param tags: list of tags of the scenario
        :return: node
        """
        tags_node = etree.Element('scenarioTags')
        for tag in tags:
            tags_node.append(etree.Element(tag.value))

        return tags_node


class LaneletXMLNode:
    @classmethod
    def create_node(cls, lanelet: Lanelet) -> etree.Element:
        """
        Create XML-Node for a Lanelet
        :param lanelet: lanelet for creating a node
        :return: node
        """
        lanelet_node = etree.Element('lanelet')
        lanelet_node.set('id', str(lanelet.lanelet_id))

        # left boundary
        left_boundary = etree.Element('leftBound')
        Pointlist.create_from_numpy_array(lanelet.left_vertices).add_points_to_node(
            left_boundary
        )

        if hasattr(lanelet, 'line_marking_left_vertices') and isinstance(
            lanelet.line_marking_left_vertices, LineMarking
        ) and lanelet.line_marking_left_vertices is not LineMarking.UNKNOWN:
            line_marking_left = etree.Element('lineMarking')
            line_marking_left.text = lanelet.line_marking_left_vertices.value
            left_boundary.append(line_marking_left)

        lanelet_node.append(left_boundary)

        # right boundary
        right_boundary = etree.Element('rightBound')
        Pointlist.create_from_numpy_array(lanelet.right_vertices).add_points_to_node(
            right_boundary
        )

        if hasattr(lanelet, 'line_marking_right_vertices') and isinstance(
            lanelet.line_marking_right_vertices, LineMarking
        ) and lanelet.line_marking_right_vertices is not LineMarking.UNKNOWN:
            line_marking_right = etree.Element('lineMarking')
            line_marking_right.text = lanelet.line_marking_right_vertices.value
            right_boundary.append(line_marking_right)

        lanelet_node.append(right_boundary)

        for l in lanelet.predecessor:
            predecessor = etree.Element('predecessor')
            predecessor.set('ref', str(l))
            lanelet_node.append(predecessor)

        for l in lanelet.successor:
            successor = etree.Element('successor')
            successor.set('ref', str(l))
            lanelet_node.append(successor)

        if lanelet.adj_left:
            adjacent_left = etree.Element('adjacentLeft')
            adjacent_left.set('ref', str(lanelet.adj_left))
            if lanelet.adj_left_same_direction:
                adjacent_left.set('drivingDir', 'same')
            else:
                adjacent_left.set('drivingDir', 'opposite')
            lanelet_node.append(adjacent_left)

        if lanelet.adj_right:
            adjacent_right = etree.Element('adjacentRight')
            adjacent_right.set('ref', str(lanelet.adj_right))
            if lanelet.adj_right_same_direction:
                adjacent_right.set('drivingDir', 'same')
            else:
                adjacent_right.set('drivingDir', 'opposite')
            lanelet_node.append(adjacent_right)

        if lanelet.stop_line:
            stop_line_node = LaneletStopLineXMLNode.create_node(lanelet.stop_line)
            lanelet_node.append(stop_line_node)

        if len(lanelet.lanelet_type) > 0:
            for lanelet_type_element in lanelet.lanelet_type:
                lanelet_type_node = etree.Element('laneletType')
                lanelet_type_node.text = str(lanelet_type_element.value)
                lanelet_node.append(lanelet_type_node)
        else:
            warnings.warn('<CommonRoadFileWriter/lanelet.lanelet_type> Lanelet %s has no '
                          'lanelet type! Default lanelet type is used!' % lanelet.lanelet_id)
            lanelet_type_node = etree.Element('laneletType')
            lanelet_type_node.text = str(LaneletType.UNKNOWN.value)
            lanelet_node.append(lanelet_type_node)

        if lanelet.user_one_way:
            for user_one_way in lanelet.user_one_way:
                user_one_way_node = etree.Element('userOneWay')
                user_one_way_node.text = str(user_one_way.value)
                lanelet_node.append(user_one_way_node)

        if lanelet.user_bidirectional:
            for user in lanelet.user_bidirectional:
                user_node = etree.Element('userBidirectional')
                user_node.text = str(user.value)
                lanelet_node.append(user_node)

        if lanelet.traffic_signs:
            for traffic_sign in lanelet.traffic_signs:
                traffic_sign_node = TrafficSignXMLNode.create_ref_node(traffic_sign)
                lanelet_node.append(traffic_sign_node)

        if lanelet.traffic_lights:
            for traffic_light in lanelet.traffic_lights:
                traffic_light_node = TrafficLightXMLNode.create_ref_node(traffic_light)
                lanelet_node.append(traffic_light_node)

        return lanelet_node


class ObstacleXMLNode:
    @classmethod
    def create_node(cls, obstacle: Union[Obstacle, DynamicObstacle, StaticObstacle, EnvironmentObstacle,
                                         PhantomObstacle]) -> etree.Element:
        """
        Create XML-Node for an Obstacle
        :param obstacle: Obstacle for creating a node
        :return:
        """
        if type(obstacle) == DynamicObstacle:
            return DynamicObstacleXMLNode.create_node(obstacle)
        elif type(obstacle) == StaticObstacle:
            return StaticObstacleXMLNode.create_node(obstacle)
        elif type(obstacle) == EnvironmentObstacle:
            return EnvironmentObstacleXMLNode.create_node(obstacle)
        elif type(obstacle) == PhantomObstacle:
            return PhantomObstacleXMLNode.create_node(obstacle)
        else:
            raise Exception()

    @classmethod
    def create_obstacle_node_header(
        cls, obstacle_id: int, obstacle_role: ObstacleRole, obstacle_type: ObstacleType
    ):
        obstacle_node = etree.Element(obstacle_role.value+'Obstacle')
        obstacle_node.set('id', str(obstacle_id))
        type_node = etree.Element('type')
        type_node.text = obstacle_type.value
        obstacle_node.append(type_node)
        return obstacle_node


class EnvironmentObstacleXMLNode:
    @classmethod
    def create_node(cls, environment_obstacle: EnvironmentObstacle) -> etree.Element:
        """
        Create XML-Node for a EnvironmentObstacle
        :param environment_obstacle: EnvironmentObstacle for creating a node
        :return: node
        """
        node = ObstacleXMLNode.create_obstacle_node_header(
            environment_obstacle.obstacle_id,
            environment_obstacle.obstacle_role,
            environment_obstacle.obstacle_type,
        )
        shape_node = etree.Element('shape')
        shape_node.extend(ShapeXMLNode.create_node(environment_obstacle.obstacle_shape))
        node.append(shape_node)

        return node


class PhantomObstacleXMLNode:
    @classmethod
    def create_node(cls, phantom_obstacle: PhantomObstacle) -> etree.Element:
        """
        Create XML-Node for a PhantomObstacle
        :param phantom_obstacle: PhantomObstacle for creating a node
        :return: node
        """
        node = PhantomObstacleXMLNode.create_obstacle_node_header(phantom_obstacle.obstacle_id,
                                                                  phantom_obstacle.obstacle_role)
        if isinstance(phantom_obstacle.prediction, SetBasedPrediction):
            node.append(DynamicObstacleXMLNode.create_occupancy_node(phantom_obstacle.prediction.occupancy_set))
        return node

    @classmethod
    def create_obstacle_node_header(cls, obstacle_id: int, obstacle_role: ObstacleRole):
        obstacle_node = etree.Element(obstacle_role.value+'Obstacle')
        obstacle_node.set('id', str(obstacle_id))
        return obstacle_node


class StaticObstacleXMLNode:
    @classmethod
    def create_node(cls, static_obstacle: StaticObstacle) -> etree.Element:
        """
        Create XML-Node for a StaticObstacle
        :param static_obstacle: static_obstacle for creating a node
        :return: node
        """
        node = ObstacleXMLNode.create_obstacle_node_header(
            static_obstacle.obstacle_id,
            static_obstacle.obstacle_role,
            static_obstacle.obstacle_type,
        )
        shape_node = etree.Element('shape')
        shape_node.extend(ShapeXMLNode.create_node(static_obstacle.obstacle_shape))
        node.append(shape_node)

        # write intial state
        initial_state_node = etree.Element('initialState')
        StateXMLNode.create_state_node(
            static_obstacle.initial_state,
            initial_state_node,
            time_step=static_obstacle.initial_state.time_step,
        )
        node.append(initial_state_node)

        return node


class DynamicObstacleXMLNode:
    @classmethod
    def create_node(cls, dynamic_obstacle: DynamicObstacle) -> etree.Element:
        """
        Create XML-Node for a DynamicObstacle
        :param dynamic_obstacle: dynamic_obstacle for creating a node
        :return: node
        """
        obstacle_node = ObstacleXMLNode.create_obstacle_node_header(
            dynamic_obstacle.obstacle_id,
            dynamic_obstacle.obstacle_role,
            dynamic_obstacle.obstacle_type,
        )
        shape_node = etree.Element('shape')
        shape_node.extend(
            ShapeXMLNode.create_node(
                dynamic_obstacle.obstacle_shape, dynamic_obstacle_shape=True
            )
        )
        obstacle_node.append(shape_node)

        # write intial state
        initial_state_node = etree.Element('initialState')
        StateXMLNode.create_state_node(
            dynamic_obstacle.initial_state,
            initial_state_node,
            time_step=dynamic_obstacle.initial_state.time_step,
        )
        obstacle_node.append(initial_state_node)

        # write initial signal state if one exists
        if dynamic_obstacle.initial_signal_state is not None:
            initial_signal_state_node = etree.Element('initialSignalState')
            SignalStateXMLNode.create_signal_state_node(
                dynamic_obstacle.initial_signal_state,
                initial_signal_state_node,
                time_step=dynamic_obstacle.initial_signal_state.time_step,
            )
            obstacle_node.append(initial_signal_state_node)

        # write prediction depending on type
        if isinstance(dynamic_obstacle.prediction, SetBasedPrediction):
            obstacle_node.append(
                cls.create_occupancy_node(dynamic_obstacle.prediction.occupancy_set)
            )
        elif isinstance(dynamic_obstacle.prediction, TrajectoryPrediction):
            obstacle_node.append(
                cls._create_trajectory_node(dynamic_obstacle.prediction.trajectory)
            )

        # write signal series if it exists
        if dynamic_obstacle.signal_series is not None and len(dynamic_obstacle.signal_series) > 0:
            obstacle_node.append(cls._create_signal_series_node(dynamic_obstacle.signal_series))

        return obstacle_node

    @classmethod
    def _create_trajectory_node(cls, trajectory: Trajectory) -> etree.Element:
        """
        Create XML-Node for a Trajectory
        :param trajectory: trajectory for creating a node
        :return: node
        """
        traj_node = etree.Element('trajectory')
        for state in trajectory.state_list:
            state_node = etree.Element('state')
            traj_node.append(
                StateXMLNode.create_state_node(state, state_node, state.time_step)
            )
        return traj_node

    @classmethod
    def create_occupancy_node(cls, occupancy_set: List[Occupancy]) -> etree.Element:
        """
        Create XML-Node for an occupancy_set
        :param occupancy_set: occupancy_set for creating a node
        :return: node
        """
        occupancy_set_node = etree.Element('occupancySet')
        for occupancy in occupancy_set:
            occupancy_set_node.append(OccupancyXMLNode.create_node(occupancy))
        return occupancy_set_node

    @classmethod
    def _create_signal_series_node(cls, signal_series: List[SignalState]) -> etree.Element:
        """
        Create XML-Node for a Trajectory
        :param trajectory: trajectory for creating a node
        :return: node
        """
        series_node = etree.Element('signalSeries')
        for signal_state in signal_series:
            signal_state_node = etree.Element('signalState')
            series_node.append(
                SignalStateXMLNode.create_signal_state_node(signal_state, signal_state_node, signal_state.time_step)
            )
        return series_node


class OccupancyXMLNode:
    @classmethod
    def create_node(cls, occupancy: Occupancy) -> etree.Element:
        """
        Create XML-Node for an Occupancy
        :param occupancy: occupancy for creating a node
        :return: node
        """
        occupancy_node = etree.Element('occupancy')

        shape_node = etree.Element('shape')
        shape_node.extend(ShapeXMLNode.create_node(occupancy.shape))
        occupancy_node.append(shape_node)

        time_node = etree.Element('time')
        time = occupancy.time_step
        if isinstance(occupancy.time_step, Interval):
            time_node.extend(create_interval_node_int(time))
        else:
            time_node.append(create_exact_node_int(time))
        occupancy_node.append(time_node)

        return occupancy_node


class ShapeXMLNode:
    @classmethod
    def create_node(cls, shape, dynamic_obstacle_shape=False) -> List[etree.Element]:
        """
        Create XML-Node for a shape
        :param shape: shape for creating a node
        :param dynamic_obstacle_shape: specify whether the shape belongs to an dynamic obstacle or not
        :return: node
        """
        if type(shape) == ShapeGroup:
            shape_node_list = []
            for s in shape.shapes:
                shape_node_list.append(
                    cls._create_single_element(s, dynamic_obstacle_shape)
                )
        else:
            shape_node = cls._create_single_element(shape, dynamic_obstacle_shape)
            shape_node_list = [shape_node]
        return shape_node_list

    @classmethod
    def _create_single_element(
        cls, shape: Shape, dynamic_obstacle_shape: bool
    ) -> etree.Element:
        """
        Create XML-Node for a single shape element
        :param shape: shape for creating a node
        :param dynamic_obstacle_shape: specify whether the shape belongs to an dynamic obstacle or not
        :return: node
        """
        if type(shape) == Rectangle:
            node = RectangleXMLNode.create_rectangle_node(shape, dynamic_obstacle_shape)
        elif type(shape) == Circle:
            node = CircleXMLNode.create_circle_node(shape, dynamic_obstacle_shape)
        elif type(shape) == Polygon:
            node = PolygonXMLNode.create_polygon_node(shape, dynamic_obstacle_shape)
        else:
            raise TypeError(
                '<ShapeXMLNode/_create_single_element> Expected type Polygon, Circle or Rectangle but got %s'
                % (type(shape))
            )
        return node


class RectangleXMLNode:
    @classmethod
    def create_rectangle_node(
        cls, rectangle: Rectangle, dynamic_obstacle_shape=False
    ) -> etree.Element:
        """
        Create XML-Node for a rectangle
        :param rectangle: rectangle for creating a node
        :param dynamic_obstacle_shape: specify whether the shape belongs to an dynamic obstacle or not
        :return: node
        """
        rectangle_node = etree.Element('rectangle')
        length_node = etree.Element('length')
        length_node.text = str(rectangle.length)
        rectangle_node.append(length_node)

        width_node = etree.Element('width')
        width_node.text = str(rectangle.width)
        rectangle_node.append(width_node)

        if not dynamic_obstacle_shape:
            orientation_node = etree.Element('orientation')
            orientation_node.text = str(np.float64(rectangle.orientation))
            rectangle_node.append(orientation_node)

            center_node = etree.Element('center')
            x_node = etree.Element('x')
            x_node.text = str(np.float64(rectangle.center[0]))
            center_node.append(x_node)
            y_node = etree.Element('y')
            y_node.text = str(np.float64(rectangle.center[1]))
            center_node.append(y_node)
            rectangle_node.append(center_node)
        return rectangle_node


class CircleXMLNode:
    @classmethod
    def create_circle_node(
        cls, circle: Circle, dynamic_obstacle_shape=False
    ) -> etree.Element:
        """
        Create XML-Node for a circle
        :param circle: circle for creating a node
        :param dynamic_obstacle_shape: specify whether the shape belongs to an dynamic obstacle or not
        :return: node
        """
        circle_node = etree.Element('circle')

        radius_node = etree.Element('radius')
        radius_node.text = str(np.float64(circle.radius))
        circle_node.append(radius_node)

        if not dynamic_obstacle_shape:
            center_node = etree.Element('center')
            x_node = etree.Element('x')
            x_node.text = str(np.float64(circle.center[0]))
            center_node.append(x_node)
            y_node = etree.Element('y')
            y_node.text = str(np.float64(circle.center[1]))
            center_node.append(y_node)
            circle_node.append(center_node)
        return circle_node


class PolygonXMLNode:
    @classmethod
    def create_polygon_node(
        cls, polygon: Polygon, dynamic_obstacle_shape: bool = False
    ) -> etree.Element:
        """
        Create XML-Node for a polygon
        :param polygon: polygon for creating a node
        :param dynamic_obstacle_shape: specify whether the shape belongs to an dynamic obstacle or not
        :return: node
        """
        polygon_node = etree.Element('polygon')
        for p in polygon.vertices:
            polygon_node.append(Point(p[0], p[1]).create_node())
        return polygon_node


class StateXMLNode:
    @classmethod
    def create_goal_state_node(cls, state: State, goal_lanelet_ids: List[int]) -> etree.Element:
        """
        Create XML-Node for a state
        :param state: CommonRoad state
        :param goal_lanelet_ids: contains a list of lanelet ids if a goal state's position is specified lanelet id(s)
        :return: node
        """
        state_node = etree.Element('goalState')
        if hasattr(state, 'position') or len(goal_lanelet_ids) > 0:
            position = etree.Element('position')
            position = cls._write_goal_position(position, state.position, goal_lanelet_ids)
            state_node.append(position)

        if hasattr(state, 'orientation'):
            orientation = etree.Element('orientation')
            orientation = cls._write_value_exact_or_interval(
                orientation, state.orientation
            )
            state_node.append(orientation)
        if hasattr(state, 'time_step'):
            time = etree.Element('time')
            time = cls._write_goal_time_exact_or_interval(time, state.time_step)
            state_node.append(time)
        if hasattr(state, 'velocity'):
            velocity = etree.Element('velocity')
            velocity = cls._write_value_exact_or_interval(velocity, state.velocity)
            state_node.append(velocity)
        if hasattr(state, 'acceleration'):
            acceleration = etree.Element('acceleration')
            acceleration = cls._write_value_exact_or_interval(
                acceleration, state.acceleration
            )
            state_node.append(acceleration)
        if hasattr(state, 'yaw_rate'):
            yaw_rate = etree.Element('yawRate')
            yaw_rate = cls._write_value_exact_or_interval(yaw_rate, state.yaw_rate)
            state_node.append(yaw_rate)
        if hasattr(state, 'slip_angle'):
            slip_angle = etree.Element('slipAngle')
            slip_angle = cls._write_value_exact_or_interval(
                slip_angle, state.slip_angle
            )
            state_node.append(slip_angle)
        return state_node

    @classmethod
    def _write_goal_position(
        cls, node: etree.Element, position: Union[Shape, int, list], goal_lanelet_ids: List[int],
    ) -> etree.Element:
        """
        Create XML-Node for a goal position
        :param node: node of the GoalState
        :param position: either (list of) shape elements or lanelet ids specifying the goal position
        :return: node
        """
        if len(goal_lanelet_ids) > 0:
            for id in goal_lanelet_ids:
                lanelet = etree.Element('lanelet')
                lanelet.set('ref', str(id))
                node.append(lanelet)
        elif isinstance(position, int):
            lanelet = etree.Element('lanelet')
            lanelet.set('ref', str(position))
            node.append(lanelet)
        elif(
            isinstance(position, Rectangle)
            or isinstance(position, Circle)
            or isinstance(position, Polygon)
            ):
            node.extend(ShapeXMLNode.create_node(position))
        elif isinstance(position, ShapeGroup):
            node.extend(ShapeXMLNode.create_node(position))
        elif type(position) is list:
            raise ValueError('A goal state cannot contain multiple items. Use a list of goal states instead.')
        else:
            raise ValueError('Case should not occur, position={}, goal_lanelet_ids={}.'.format(position,
                                                                                               goal_lanelet_ids))
        return node

    @classmethod
    def _write_goal_time_exact_or_interval(
        cls, node: etree.Element, time_step: Union[Interval, float, int]
    ) -> etree.Element:
        """
        Create XML-Node for a goal time
        :param node: node of the GoalState
        :param time_step: contains time interval or time_step of goal time
        :return: node
        """
        if isinstance(time_step, int):
            node.append(create_exact_node_int(time_step))
        elif isinstance(time_step, Interval):
            node.extend(
                create_interval_node_int(Interval(time_step.start, time_step.end))
            )
        else:
            raise Exception()
        return node

    @classmethod
    def _write_value_exact_or_interval(
        cls, node: etree.Element, var: Union[Interval, float, int]
    ):
        """
        Create XML-Node for a goal value
        :param node: node of the GoalState
        :param var: contains interval or exact_value of goal state value
        :return: node
        """
        if isinstance(var, (float, int)):
            node.append(create_exact_node_float(var))
        elif isinstance(var, Interval):
            node.extend(create_interval_node_float(var))
        else:
            raise Exception()
        return node

    @classmethod
    def create_state_node(
        cls, state: State, state_node: etree.Element, time_step: int
    ) -> etree.Element:
        """
        Create XML-Node for a state
        :param state: value of the state
        :param state_node: node of the overlying state
        :return: node
        """

        if hasattr(state, 'position'):
            position = etree.Element('position')
            if type(state.position) in [np.ndarray, list]:
                position.append(
                    Point.create_from_numpy_array(state.position).create_node()
                )
                state_node.append(position)
            elif isinstance(state.position, Shape):
                position.extend(ShapeXMLNode.create_node(state.position))
                state_node.append(position)
            else:
                raise Exception()
        if hasattr(state, 'orientation'):
            orientation = etree.Element('orientation')
            orientation = cls._write_value_exact_or_interval(
                orientation, state.orientation
            )
            state_node.append(orientation)

        time_node = etree.Element('time')
        time_node.append(create_exact_node_int(time_step))
        state_node.append(time_node)

        if hasattr(state, 'velocity'):
            velocity = etree.Element('velocity')
            velocity = cls._write_value_exact_or_interval(velocity, state.velocity)
            state_node.append(velocity)

        if hasattr(state, 'acceleration'):
            acceleration = etree.Element('acceleration')
            acceleration = cls._write_value_exact_or_interval(
                acceleration, state.acceleration
            )
            state_node.append(acceleration)
        if hasattr(state, 'yaw_rate'):
            yaw_rate = etree.Element('yawRate')
            yaw_rate = cls._write_value_exact_or_interval(yaw_rate, state.yaw_rate)
            state_node.append(yaw_rate)
        if hasattr(state, 'slip_angle'):
            slip_angle = etree.Element('slipAngle')
            slip_angle = cls._write_value_exact_or_interval(
                slip_angle, state.slip_angle
            )
            state_node.append(slip_angle)
        return state_node


class PlanningProblemXMLNode:
    @classmethod
    def create_node(cls, planning_problem: PlanningProblem) -> etree.Element:
        """
        Create a xml-Node for a single planning_problem
        :param planning_problem: planning problem for creating the node
        :return:
        """
        planning_problem_node = etree.Element('planningProblem')
        planning_problem_node.set('id', str(planning_problem.planning_problem_id))
        initial_state_node = etree.Element('initialState')
        planning_problem_node.append(
            StateXMLNode.create_state_node(
                planning_problem.initial_state,
                initial_state_node,
                planning_problem.initial_state.time_step,
            )
        )

        for state_id, goal_state in enumerate(planning_problem.goal.state_list):
            if (
                planning_problem.goal.lanelets_of_goal_position is not None
                and state_id in planning_problem.goal.lanelets_of_goal_position
            ):
                goal_lanelet_ids: List[
                    int
                ] = planning_problem.goal.lanelets_of_goal_position[state_id]
            else:
                goal_lanelet_ids = []

            planning_problem_node.append(
                StateXMLNode.create_goal_state_node(goal_state, goal_lanelet_ids)
            )

        return planning_problem_node


class Point:
    def __init__(self, x: Union[int, float], y: Union[int, float], z: Union[int, float, None] = None):
        self.x: Union[int, float] = x
        self.y: Union[int, float] = y
        self.z: Union[int, float] = z

    def as_numpy_array(self):
        if self.z is None:
            return np.array([self.x, self.y])
        else:
            return np.array([self.x, self.y, self.z])

    @classmethod
    def create_from_numpy_array(cls, point: Union[np.array, list]):
        assert 2 <= len(point) <= 3
        if len(point) == 2:
            return cls(point[0], point[1])
        else:
            return cls(point[0], point[1], point[2])

    def create_node(self):
        point_node = etree.Element('point')
        x = etree.Element('x')
        x.text = float_to_str(np.float64(self.x))
        point_node.append(x)
        y = etree.Element('y')
        y.text = float_to_str(np.float64(self.y))
        point_node.append(y)
        if self.z is not None:
            z = etree.Element('z')
            z.text = float_to_str(np.float64(self.z))
            point_node.append(z)
        return point_node


class Pointlist:
    def __init__(self, points: List[Point]):
        self.points = points

    @classmethod
    def create_from_numpy_array(cls, points: np.array):
        point_list = []
        for point in points:
            point_list.append(Point.create_from_numpy_array(point))
        return cls(point_list)

    def add_points_to_node(self, xml_node: etree.Element):
        for point in self.points:
            xml_node.append(point.create_node())


class IntersectionXMLNode:
    @classmethod
    def create_node(cls, intersection: Intersection) -> etree.Element:
        """
        Create a xml node for an intersection
        :param intersection: intersection for the node
        :return:
        """
        intersection_node = etree.Element('intersection')
        intersection_node.set('id', str(intersection.intersection_id))

        for incoming in intersection.incomings:
            incoming_node = etree.Element('incoming')
            incoming_node.set('id', str(incoming.incoming_id))
            for incoming_lanelet in incoming.incoming_lanelets:
                incoming_lanelet_node = etree.Element('incomingLanelet')
                incoming_lanelet_node.set('ref', str(incoming_lanelet))
                incoming_node.append(incoming_lanelet_node)

            if incoming.successors_right:
                for successor_right in incoming.successors_right:
                    successor_right_node = etree.Element('successorsRight')
                    successor_right_node.set('ref', str(successor_right))
                    incoming_node.append(successor_right_node)

            if incoming.successors_straight:
                for successor_straight in incoming.successors_straight:
                    successor_straight_node = etree.Element('successorsStraight')
                    successor_straight_node.set('ref', str(successor_straight))
                    incoming_node.append(successor_straight_node)

            if incoming.successors_left:
                for successor_left in incoming.successors_left:
                    successor_left_node = etree.Element('successorsLeft')
                    successor_left_node.set('ref', str(successor_left))
                    incoming_node.append(successor_left_node)

            if incoming.left_of:
                is_left_of_node = etree.Element('isLeftOf')
                is_left_of_node.set('ref', str(incoming.left_of))
                incoming_node.append(is_left_of_node)

            intersection_node.append(incoming_node)

        if intersection.crossings is not None and len(intersection.crossings) > 0:
            crossing_node = etree.Element('crossing')
            for crossing_lanelet in intersection.crossings:
                crossing_lanelet_node = etree.Element('crossingLanelet')
                crossing_lanelet_node.set('ref', str(crossing_lanelet))
                crossing_node.append(crossing_lanelet_node)
            intersection_node.append(crossing_node)

        return intersection_node


class TrafficSignXMLNode:
    @classmethod
    def create_node(cls, traffic_sign: TrafficSign) -> etree.Element:
        traffic_sign_node = etree.Element('trafficSign')
        traffic_sign_node.set('id', str(traffic_sign.traffic_sign_id))
        for element in traffic_sign.traffic_sign_elements:
            element_node = etree.Element('trafficSignElement')
            sign_id_node = etree.Element('trafficSignID')
            sign_id_node.text = str(element.traffic_sign_element_id.value)
            if str(element.traffic_sign_element_id.value) == '':
                warnings.warn('<FileWriter>: Invalid traffic sign ID!')
            element_node.append(sign_id_node)
            for value in element.additional_values:
                value_node = etree.Element('additionalValue')
                value_node.text = str(value)
                element_node.append(value_node)
            traffic_sign_node.append(element_node)

        if traffic_sign.position is not None:
            position_node = etree.Element('position')
            position_node.append(Point(traffic_sign.position[0],
                                       traffic_sign.position[1]).create_node())
            traffic_sign_node.append(position_node)

        if traffic_sign.virtual is not None:
            virtual_node = etree.Element('virtual')
            virtual_node.text = str(traffic_sign.virtual).lower()
            traffic_sign_node.append(virtual_node)
        return traffic_sign_node

    @classmethod
    def create_ref_node(cls, traffic_sign_ref) -> etree.Element:
        traffic_sign_ref_node = etree.Element('trafficSignRef')
        traffic_sign_ref_node.set('ref', str(traffic_sign_ref))
        return traffic_sign_ref_node


class TrafficLightXMLNode:
    @classmethod
    def create_node(cls, traffic_light: TrafficLight) -> etree.Element:
        traffic_light_node = etree.Element('trafficLight')
        traffic_light_node.set('id', str(traffic_light.traffic_light_id))
        cycle_node = etree.Element('cycle')
        for state in traffic_light.cycle:
            element_node = TrafficLightCycleElementXMLNode.create_node(state)
            cycle_node.append(element_node)
        if traffic_light.time_offset is not None and traffic_light.time_offset > 0:
            offset_node = etree.Element('timeOffset')
            offset_node.text = str(traffic_light.time_offset)
            cycle_node.append(offset_node)
        traffic_light_node.append(cycle_node)

        if traffic_light.position is not None:
            position_node = etree.Element('position')
            position_node.append(Point(traffic_light.position[0],
                                       traffic_light.position[1]).create_node())
            traffic_light_node.append(position_node)

        if traffic_light.direction is not TrafficLightDirection.ALL:
            direction_node = etree.Element('direction')
            direction_node.text = traffic_light.direction.value
            traffic_light_node.append(direction_node)

        if traffic_light.active is not None:
            active_node = etree.Element('active')
            active_node.text = str(traffic_light.active).lower()
            traffic_light_node.append(active_node)

        return traffic_light_node

    @classmethod
    def create_ref_node(cls, traffic_light_ref) -> etree.Element:
        traffic_light_ref_node = etree.Element('trafficLightRef')
        traffic_light_ref_node.set('ref', str(traffic_light_ref))
        return traffic_light_ref_node


class TrafficLightCycleElementXMLNode:
    @classmethod
    def create_node(cls, cycle_element: TrafficLightCycleElement) -> etree.Element:
        element_node = etree.Element("cycleElement")
        duration_node = etree.Element("duration")
        duration_node.text = str(cycle_element.duration)
        element_node.append(duration_node)
        color_node = etree.Element("color")
        color_node.text = cycle_element.state.value
        element_node.append(color_node)

        return element_node


class LineMarkingXMLNode:
    @classmethod
    def _line_marking_enum_to_string(cls, line_marking):
        return str(line_marking.name.lower())

    @classmethod
    def create_node(cls, line_marking: LineMarking) -> etree.Element:
        line_marking_node = etree.Element('lineMarking')
        line_marking_node.text = cls._line_marking_enum_to_string(line_marking)
        return line_marking_node


class LaneletStopLineXMLNode:
    @classmethod
    def create_node(cls, stop_line: StopLine) -> etree.Element:
        stop_line_node = etree.Element('stopLine')

        if stop_line.start is not None or stop_line.end is not None:
            start_node = Point(stop_line.start[0], stop_line.start[1]).create_node()
            stop_line_node.append(start_node)
            end_node = Point(stop_line.end[0], stop_line.end[1]).create_node()
            stop_line_node.append(end_node)

        if stop_line.line_marking:
            line_marking_node = LineMarkingXMLNode.create_node(stop_line.line_marking)
            stop_line_node.append(line_marking_node)

        if stop_line.traffic_sign_ref is not None:
            for sign in stop_line.traffic_sign_ref:
                traffic_sign_ref_node = TrafficSignXMLNode.create_ref_node(sign)
                stop_line_node.append(traffic_sign_ref_node)

        if stop_line.traffic_light_ref is not None:
            for light in stop_line.traffic_light_ref:
                traffic_light_ref_node = TrafficLightXMLNode.create_ref_node(light)
                stop_line_node.append(traffic_light_ref_node)

        return stop_line_node


class SignalStateXMLNode:
    @classmethod
    def create_signal_state_node(cls, signal_state: SignalState, signal_state_node: etree.Element,
                          time_step: int) -> etree.Element:
        """
        Create XML-Node for a state
        :param signal_state: value of the signal state
        :param signal_state_node: node of the overlying state
        :return: node
        """
        time_node = etree.Element('time')
        time_node.append(create_exact_node_int(time_step))
        signal_state_node.append(time_node)

        if hasattr(signal_state, 'indicator_left'):
            indicator_left = etree.Element('indicatorLeft')
            indicator_left.text = str(signal_state.indicator_left).lower()
            signal_state_node.append(indicator_left)

        if hasattr(signal_state, 'indicator_right'):
            indicator_right = etree.Element('indicatorRight')
            indicator_right.text = str(signal_state.indicator_right).lower()
            signal_state_node.append(indicator_right)

        if hasattr(signal_state, 'braking_lights'):
            braking_lights = etree.Element('brakingLights')
            braking_lights.text = str(signal_state.braking_lights).lower()
            signal_state_node.append(braking_lights)

        if hasattr(signal_state, 'hazard_warning_lights'):
            hazard_warning_lights = etree.Element('hazardWarningLights')
            hazard_warning_lights.text = str(signal_state.braking_lights).lower()
            signal_state_node.append(hazard_warning_lights)

        if hasattr(signal_state, 'flashing_blue_lights'):
            flashing_blue_lights = etree.Element('flashingBlueLights')
            flashing_blue_lights.text = str(signal_state.flashing_blue_lights).lower()
            signal_state_node.append(flashing_blue_lights)

        return signal_state_node



import json
import logging
import os
from typing import Tuple, Union

# Read defaults only once
default_params_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   'default_draw_params.json')
with open(default_params_path) as fp:
    default_params = json.load(fp)


def write_default_params(filename: str) -> None:
    """
    Write default parameters to the give file as JSON

    :param filename: Destination JSON file
    :return: None
    """
    with open(filename, 'w') as fp:
        json.dump(default_params, fp, indent=4)


class ParamServer:
    """
    Wrapper object for parameters as nested dictionaries. Tries to resolve
    queries with most specialized values. If values are not found, defaults
    are used.
    """

    def __init__(self, params: dict = None, warn_default: bool = False, default=None):
        """
        :param params: Optional parameters to initialize parameter server with
        :param warn_default: Produce a warning when default parameters are used
        :param default: Optional default parameter set. If provided, overrides the defaults in default_draw_params.json.
        """
        self._params = params or {}
        self._warn_default = warn_default
        if isinstance(default, dict):
            self._default = ParamServer(params=default)
        elif isinstance(default, ParamServer):
            self._default = default
        else:
            self._default = default_params

    @staticmethod
    def _resolve_key(param_dict, key):
        if isinstance(param_dict, ParamServer):
            return param_dict.resolve_key(key)
        else:
            if len(key) == 0:
                return None, 0
            tmp_dict = param_dict
            l_key = list(key)
            # Try to find most special version of element
            while len(l_key) > 0:
                k = l_key.pop(0)
                if k in tmp_dict.keys():
                    tmp_dict = tmp_dict[k]
                else:
                    tmp_dict = None
                    break
            if tmp_dict is None and len(key) > 0:
                # If not found, remove first level and try again
                return ParamServer._resolve_key(param_dict, key[1:])
            else:
                return tmp_dict, len(key)

    def resolve_key(self, param_path):
        val, depth = ParamServer._resolve_key(self._params, param_path)
        val_default, depth_default = ParamServer._resolve_key(self._default, param_path)
        if val is None and val_default is None:
            return None, None

        if val_default is not None and val is None:
            if self._warn_default:
                logging.warning('Using default for key {}!'.format(param_path))
            return val_default, depth_default

        if val is not None and val_default is None:
            return val, depth

        if val is not None and val_default is not None:
            if depth >= depth_default:
                return val, depth
            else:
                return val_default, depth_default

    def by_callstack(self, call_stack: Tuple[str, ...], param_path: Union[str, Tuple[str, ...]]):
        """
        Resolves the parameter path using the callstack. If it nothing can be
        found returns None

        :param call_stack: Tuple of string containing the call stack,
            which allows for differentiation of plotting styles depending on the call stack
        :param param_path: Key or tuple of keys leading to the parameter
        :return: the parameter
        """
        if isinstance(param_path, tuple):
            path = call_stack + param_path
        else:
            path = call_stack + (param_path,)
        return self.__getitem__(path)

    def __getitem__(self, param_path):
        """
        Resolves the parameter by the given key tuple. Parameters are
        resolved recursively. If no parameter can be found under the given
        path, the first element of the tuple is removed and the resolution
        will be retried. This yields the most specialized version of the
        parameter. Default parameters are provided if:

        a) the specified path cannot be resolved in the contained parameters or
        b) the default parameters contain a more specific version than the
        contained parameters

        :param param_path: Key or tuple of keys leading to the parameter
        :return: the parameter
        """
        if not isinstance(param_path, tuple):
            param_path = (param_path,)

        val, _ = self.resolve_key(param_path)
        if val is None:
            logging.error('Value for key {} not found!'.format(param_path))
        return val

    def __setitem__(self, param_path, value):
        """
        Sets the value under the given key

        :param param_path: key or tuple of keys leading to the parameter
        :param value: the value
        :return: None
        """
        if not isinstance(param_path, tuple):
            param_path = (param_path,)
        tmp_dict = self._params
        for key in param_path[:-1]:
            if not isinstance(tmp_dict, dict):
                raise KeyError(
                        'Key "{}" in path "{}" is not subscriptable!'.format(
                            key, param_path))
            if key in tmp_dict.keys():
                tmp_dict = tmp_dict[key]
            else:
                tmp_dict[key] = {}
                tmp_dict = tmp_dict[key]
        if not isinstance(tmp_dict, dict):
            raise KeyError(
                    'Key "{}" in path "{}" is not subscriptable!'.format(key,
                                                                         param_path))
        tmp_dict[param_path[-1]] = value

    def __contains__(self, item):
        return item in self._params

    def update(self, source):
        self._update_recurisve(self._params, source)

    def _update_recurisve(self, dest, source):
        for k, v in source.items():
            if k in dest:
                if isinstance(dest[k], dict):
                    assert isinstance(source[k], dict)
                    self._update_recurisve(dest[k], source[k])
                else:
                    dest[k] = source[k]
            else:
                dest[k] = source[k]

    @staticmethod
    def from_json(fname: str):
        """
        Restores a parameter server from a JSON file

        :param fname: file name and path of the JSON file
        :return: the parameter server
        """
        with open(fname, 'r') as fp:
            data = json.load(fp)
        return ParamServer(data)



from collections import defaultdict
from typing import Dict
from xml.etree import ElementTree
from abc import ABC

from commonroad import SUPPORTED_COMMONROAD_VERSIONS
from commonroad.common.util import Interval, AngleInterval
from commonroad.geometry.shape import Rectangle, Circle, Polygon, ShapeGroup, Shape
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem
from commonroad.prediction.prediction import Occupancy, SetBasedPrediction, TrajectoryPrediction
from commonroad.scenario.lanelet import Lanelet, LaneletNetwork, LineMarking, LaneletType, RoadUser, StopLine
from commonroad.scenario.obstacle import ObstacleType, StaticObstacle, DynamicObstacle, Obstacle, EnvironmentObstacle, \
    SignalState, PhantomObstacle
from commonroad.scenario.scenario import Scenario, Tag, GeoTransformation, Location, Environment, Time, \
    TimeOfDay, Weather, Underground, ScenarioID
from commonroad.scenario.trajectory import State, Trajectory
from commonroad.scenario.traffic_sign import *
from commonroad.scenario.intersection import Intersection, IntersectionIncomingElement


__author__ = "Stefanie Manzinger, Sebastian Maierhofer"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles", "CAR@TUM"]
__version__ = "2021.2"
__maintainer__ = "Stefanie Manzinger, Sebastian Maierhofer"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"


def read_value_exact_or_interval(xml_node: ElementTree.Element)\
        -> Union[float, Interval]:
    if xml_node.find('exact') is not None:
        value = float(xml_node.find('exact').text)
    elif xml_node.find('intervalStart') is not None \
            and xml_node.find('intervalEnd') is not None:
        value = Interval(
            float(xml_node.find('intervalStart').text),
            float(xml_node.find('intervalEnd').text))
    else:
        raise Exception()
    return value


def read_time(xml_node: ElementTree.Element) -> Union[int, Interval]:
    if xml_node.find('exact') is not None:
        value = int(xml_node.find('exact').text)
    elif xml_node.find('intervalStart') is not None \
            and xml_node.find('intervalEnd') is not None:
        value = Interval(
            int(xml_node.find('intervalStart').text),
            int(xml_node.find('intervalEnd').text))
    else:
        raise Exception()
    return value


class CommonRoadFileReader:
    """ Class which reads CommonRoad XML-files. The XML-files are composed of
    (1) a formal representation of the road network,
    (2) static and dynamic obstacles,
    (3) the planning problem of the ego vehicle(s). """
    def __init__(self, filename: str):
        """
        :param filename: full path + filename of the CommonRoad XML-file,
        """
        self._filename = filename
        self._tree = None
        self._dt = None
        self._benchmark_id = None
        self._meta_data = None

    def open(self, lanelet_assignment: bool = False) -> Tuple[Scenario, PlanningProblemSet]:
        """
        Reads a CommonRoad XML-file.

        :param lanelet_assignment: activates calculation of lanelets occupied by obstacles
        :return: the scenario containing the road network and the obstacles and the planning problem set \
        containing the planning problems---initial states and goal regions--for all ego vehicles.
        """
        self._read_header()
        scenario = self._open_scenario(lanelet_assignment)
        planning_problem_set = self._open_planning_problem_set(scenario.lanelet_network)
        return scenario, planning_problem_set

    def open_lanelet_network(self) -> LaneletNetwork:
        """
        Reads the lanelet network of a CommonRoad XML-file.

        :return: object of class LaneletNetwork
        """
        self._read_header()
        return LaneletNetworkFactory.create_from_xml_node(self._tree)

    def _open_scenario(self, lanelet_assignment: bool) -> Scenario:
        """
        Reads the lanelet network and obstacles from the CommonRoad XML-file.

        :param lanelet_assignment: activates calculation of lanelets occupied by obstacles
        :return: object of class scenario containing the road network and the obstacles
        """
        scenario = ScenarioFactory.create_from_xml_node(self._tree, self._dt, self._benchmark_id,
                                                        self._commonroad_version, self._meta_data, lanelet_assignment)
        return scenario

    def _open_planning_problem_set(self, lanelet_network: LaneletNetwork) \
            -> PlanningProblemSet:
        """
        Reads all planning problems from the CommonRoad XML-file.

        :return: object of class PlanningProblemSet containing the planning problems for all ego vehicles.
        """
        planning_problem_set = PlanningProblemSetFactory.create_from_xml_node(
            self._tree, lanelet_network)
        return planning_problem_set

    def _read_header(self):
        """ Parses the CommonRoad XML-file into element tree; reads the global time step size of the time-discrete
        scenario and the CommonRoad benchmark ID."""
        self._parse_file()
        commonroad_version = self._get_commonroad_version()
        assert commonroad_version in SUPPORTED_COMMONROAD_VERSIONS, '<CommonRoadFileReader/_read_header>: ' \
                                                                    'CommonRoad version of XML-file {} is not ' \
                                                                    'supported. Supported versions: {}. Got ' \
                                                                    'version: {}.'.format(self._filename,
                                                                                          SUPPORTED_COMMONROAD_VERSIONS,
                                                                                          commonroad_version)
        self._dt = self._get_dt()
        self._benchmark_id = self._get_benchmark_id()
        self._commonroad_version = commonroad_version
        if commonroad_version == '2018b':
            self._meta_data = {'author': self._get_author(),
                               'affiliation': self._get_affiliation(),
                               'source': self._get_source(),
                               'tags': self._get_tags(),
                               'location': Location()}
        else:
            self._meta_data = {'author': self._get_author(),
                               'affiliation': self._get_affiliation(),
                               'source': self._get_source()}

    def _parse_file(self):
        """ Parses the CommonRoad XML-file into element tree."""
        self._tree = ElementTree.parse(self._filename)

    def _get_dt(self) -> float:
        """ Reads the time step size of the time-discrete scenario."""
        return float(self._tree.getroot().get('timeStepSize'))

    def _get_benchmark_id(self) -> str:
        """ Reads the unique CommonRoad benchmark ID of the scenario."""
        return self._tree.getroot().get('benchmarkID')

    def _get_commonroad_version(self) -> str:
        """ Reads the CommonRoad version of the XML-file."""
        return self._tree.getroot().get('commonRoadVersion')

    def _get_author(self) -> str:
        """ Reads the author of the scenario."""
        return self._tree.getroot().get('author')

    def _get_affiliation(self) -> str:
        """ Reads the affiliation of the author of the scenario."""
        return self._tree.getroot().get('affiliation')

    def _get_source(self) -> str:
        """ Reads the source of the scenario."""
        return self._tree.getroot().get('source')

    def _get_tags(self) -> Set[Tag]:
        """ Reads the tags of the scenario."""
        tags_string = self._tree.getroot().get('tags')
        splits = tags_string.split()
        tags = set()
        for tag in splits:
            try:
                tags.add(Tag(tag))
            except ValueError:
                warnings.warn('Scenario tag \'{}\' not valid.'.format(tag), stacklevel=2)

        return tags


class ScenarioFactory:
    """ Class to create an object of class Scenario from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, dt: float, benchmark_id: str, commonroad_version: str,
                             meta_data: dict, lanelet_assignment: bool):
        """
        :param xml_node: XML element
        :param dt: time step size of the scenario
        :param benchmark_id: unique CommonRoad benchmark ID
        :param commonroad_version: CommonRoad version of the file
        :param lanelet_assignment: activates calculation of lanelets occupied by obstacles
        :return: CommonRoad scenario
        """
        if commonroad_version != '2018b':
            meta_data["tags"] = TagsFactory.create_from_xml_node(xml_node)
            meta_data["location"] = LocationFactory.create_from_xml_node(xml_node)
        else:
            LaneletFactory._speed_limits = {}

        scenario_id = ScenarioID.from_benchmark_id(benchmark_id, commonroad_version)
        scenario = Scenario(dt, scenario_id, **meta_data)

        scenario.add_objects(LaneletNetworkFactory.create_from_xml_node(xml_node))
        large_num = 10000
        if commonroad_version == '2018b':
            scenario.add_objects(cls._obstacles_2018b(xml_node, scenario.lanelet_network, lanelet_assignment))
            for key, value in LaneletFactory._speed_limits.items():
                for lanelet in value:
                    if SupportedTrafficSignCountry.GERMANY.value == scenario_id.country_id:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDGermany.MAX_SPEED, [str(key)])
                    elif SupportedTrafficSignCountry.USA.value == scenario_id.country_id:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDUsa.MAX_SPEED, [str(key)])
                    elif SupportedTrafficSignCountry.CHINA.value == scenario_id.country_id:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDChina.MAX_SPEED, [str(key)])
                    elif SupportedTrafficSignCountry.SPAIN.value == scenario_id.country_id:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDSpain.MAX_SPEED, [str(key)])
                    elif SupportedTrafficSignCountry.RUSSIA.value == scenario_id.country_id:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDRussia.MAX_SPEED, [str(key)])
                    elif SupportedTrafficSignCountry.ZAMUNDA.value == scenario_id.country_id:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDZamunda.MAX_SPEED, [str(key)])
                    else:
                        traffic_sign_element = TrafficSignElement(TrafficSignIDZamunda.MAX_SPEED, [str(key)])
                        warnings.warn("Unknown country: Default traffic sign IDs are used.")
                    traffic_sign = TrafficSign(scenario.generate_object_id() + large_num, [traffic_sign_element],
                                               {lanelet},
                                               scenario.lanelet_network.find_lanelet_by_id(lanelet).right_vertices[0])
                    scenario.add_objects(traffic_sign, {lanelet})
            LaneletFactory._speed_limits = {}
        else:
            scenario.add_objects(cls._obstacles(xml_node, scenario.lanelet_network, lanelet_assignment))

        return scenario

    @classmethod
    def _obstacles_2018b(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork,
                         lanelet_assignment: bool) -> List[Obstacle]:
        """
        Reads all obstacles specified in a CommonRoad XML-file.
        :param xml_node: XML element
        :param dt: time step size of the scenario
        :param lanelet_assignment: activates calculation of lanelets occupied by obstacles
        :return: list of static and dynamic obstacles specified in the CommonRoad XML-file
        """
        obstacles = list()
        for o in xml_node.findall('obstacle'):
            if o.find('role').text == 'static':
                obstacles.append(StaticObstacleFactory.create_from_xml_node(o, lanelet_network, lanelet_assignment))
            elif o.find('role').text == 'dynamic':
                obstacles.append(DynamicObstacleFactory.create_from_xml_node(o, lanelet_network, lanelet_assignment))
            else:
                raise ValueError('Role of obstacle is unknown. Got role: {}'.format(xml_node.find('role').text))
        return obstacles

    @classmethod
    def _obstacles(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork,
                   lanelet_assignment: bool) -> List[Obstacle]:
        """
        Reads all obstacles specified in a CommonRoad XML-file.
        :param xml_node: XML element
        :param dt: time step size of the scenario
        :param lanelet_assignment: activates calculation of lanelets occupied by obstacles
        :return: list of static and dynamic obstacles specified in the CommonRoad XML-file
        """
        obstacles = []
        for o in xml_node.findall('staticObstacle'):
            obstacles.append(StaticObstacleFactory.create_from_xml_node(o, lanelet_network, lanelet_assignment))
        for o in xml_node.findall('dynamicObstacle'):
            obstacles.append(DynamicObstacleFactory.create_from_xml_node(o, lanelet_network, lanelet_assignment))
        for o in xml_node.findall('environmentObstacle'):
            obstacles.append(EnvironmentObstacleFactory.create_from_xml_node(o))
        for o in xml_node.findall('phantomObstacle'):
            obstacles.append(PhantomObstacleFactory.create_from_xml_node(o))
        return obstacles


class TagsFactory:
    """ Class to create a tag set from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Set[Tag]:
        """
        :param xml_node: XML element
        :return: set of tags
        """
        tags = set()
        tag_element = xml_node.find('scenarioTags')
        for elem in Tag:
            if tag_element.find(elem.value) is not None:
                tags.add(elem)
        return tags


class LocationFactory:
    """ Class to create a location from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Union[Location, None]:
        """
        :param xml_node: XML element
        :return: location object
        """
        if xml_node.find('location') is not None:
            location_element = xml_node.find('location')
            geo_name_id = int(location_element.find('geoNameId').text)
            gps_latitude = float(location_element.find('gpsLatitude').text)
            gps_longitude = float(location_element.find('gpsLongitude').text)
            if location_element.find('geoTransformation') is not None:
                geo_transformation = GeoTransformationFactory.create_from_xml_node(
                    location_element.find('geoTransformation'))
            else:
                geo_transformation = None
            if location_element.find('environment') is not None:
                environment = EnvironmentFactory.create_from_xml_node(
                    location_element.find('environment'))
            else:
                environment = None

            return Location(geo_name_id, gps_latitude, gps_longitude, geo_transformation, environment)
        else:
            return None


class GeoTransformationFactory:
    """ Class to create a geotransformation object of an XML element according to the CommonRoad specification."""

    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> GeoTransformation:
        """
        :param xml_node: XML element
        :return: GeoTransformation object
        """
        geo_reference = xml_node.find('geoReference').text
        if xml_node.find('additionalTransformation') is not None:
            add_trans_node = xml_node.find('additionalTransformation')
            x_translation = float(add_trans_node.find('xTranslation').text)
            y_translation = float(add_trans_node.find('yTranslation').text)
            z_rotation = float(add_trans_node.find('zRotation').text)
            scaling = float(add_trans_node.find('scaling').text)
            return GeoTransformation(geo_reference, x_translation, y_translation, z_rotation, scaling)
        else:
            return GeoTransformation(geo_reference)


class EnvironmentFactory:
    """ Class to create a environment object of an XML element according to the CommonRoad specification."""

    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Environment:
        """
        :param xml_node: XML element
        :return: Environment object
        """
        time = TimeFactory.create_from_xml_node(xml_node.find('time').text)
        weather = Weather(xml_node.find('weather').text)
        underground = Underground(xml_node.find('underground').text)
        time_of_day = TimeOfDay(xml_node.find('timeOfDay').text)

        return Environment(time, time_of_day, weather, underground)


class TimeFactory:
    """ Class to create a time object of an XML element."""

    @classmethod
    def create_from_xml_node(cls, time_text: str) -> Time:
        """
        :param time_text: time as string
        :return: time object
        """
        hours = int(time_text[0:2])
        minutes = int(time_text[3:5])

        return Time(hours, minutes)


class LaneletNetworkFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> LaneletNetwork:
        """
        Reads all lanelets specified in a CommonRoad XML-file.
        :param xml_node: XML element
        :return: list of lanelets
        """
        lanelets = []
        for lanelet_node in xml_node.findall('lanelet'):
            lanelets.append(LaneletFactory.create_from_xml_node(lanelet_node))
        lanelet_network = LaneletNetwork.create_from_lanelet_list(lanelets)

        country = cls._find_country( xml_node)
        first_traffic_sign_occurence = cls._find_first_traffic_sign_occurence(lanelet_network)
        for traffic_sign_node in xml_node.findall('trafficSign'):
            lanelet_network.add_traffic_sign(TrafficSignFactory.create_from_xml_node(traffic_sign_node, country,
                                                                                     first_traffic_sign_occurence,
                                                                                     lanelet_network), [])

        for traffic_light_node in xml_node.findall('trafficLight'):
            lanelet_network.add_traffic_light(TrafficLightFactory.create_from_xml_node(traffic_light_node, country,
                                                                                       lanelet_network), [],)

        for intersection_node in xml_node.findall('intersection'):
            lanelet_network.add_intersection(IntersectionFactory.create_from_xml_node(intersection_node))

        return lanelet_network

    @staticmethod
    def _find_first_traffic_sign_occurence(lanelet_network: LaneletNetwork) -> Dict[int, Set[int]]:
        """
        Evaluates all lanelets if a traffic sign occurs first within it
        :param lanelet_network: CommonRoad lanelet network
        :return: list of tuples with traffic sign ID and corresponding lanelet ID
        """
        occurences = {}
        for lanelet in lanelet_network.lanelets:
            for traffic_sign in lanelet.traffic_signs:
                # create set object if none exist
                if occurences.get(traffic_sign) is None:
                    occurences[traffic_sign] = set()
                # if there exists no predecessor, current lanelet is first occurence
                if len(lanelet.predecessor) == 0:
                    occurences[traffic_sign].add(lanelet.lanelet_id)
                # if no predecessor references the traffic sign, this is the first occurence
                elif all(traffic_sign not in
                         lanelet_network.find_lanelet_by_id(pre).traffic_signs for pre in lanelet.predecessor):
                    occurences[traffic_sign].add(lanelet.lanelet_id)

        return occurences

    @staticmethod
    def _find_country(xml_node: ElementTree.Element) -> SupportedTrafficSignCountry:
        """
        Extracts country from location element
        :param xml_node: CommonRoad root xml node
        :return: supported traffic sign country enum
        """
        if xml_node._root.attrib["benchmarkID"][:2] == "C-":
            country = xml_node._root.attrib["benchmarkID"][2:5]
        else:
            country = xml_node._root.attrib["benchmarkID"][:3]

        if SupportedTrafficSignCountry.GERMANY.value == country:
            return SupportedTrafficSignCountry.GERMANY
        elif SupportedTrafficSignCountry.USA.value == country:
            return SupportedTrafficSignCountry.USA
        elif SupportedTrafficSignCountry.CHINA.value == country:
            return SupportedTrafficSignCountry.CHINA
        elif SupportedTrafficSignCountry.SPAIN.value == country:
            return SupportedTrafficSignCountry.SPAIN
        elif SupportedTrafficSignCountry.RUSSIA.value == country:
            return SupportedTrafficSignCountry.RUSSIA
        elif SupportedTrafficSignCountry.ARGENTINA.value == country:
            return SupportedTrafficSignCountry.ARGENTINA
        elif SupportedTrafficSignCountry.ITALY.value == country:
            return SupportedTrafficSignCountry.ITALY
        elif SupportedTrafficSignCountry.FRANCE.value == country:
            return SupportedTrafficSignCountry.FRANCE
        elif SupportedTrafficSignCountry.PUERTO_RICO.value == country:
            return SupportedTrafficSignCountry.PUERTO_RICO
        elif SupportedTrafficSignCountry.CROATIA.value == country:
            return SupportedTrafficSignCountry.CROATIA
        elif SupportedTrafficSignCountry.GREECE.value == country:
            return SupportedTrafficSignCountry.GREECE
        elif SupportedTrafficSignCountry.BELGIUM.value == country:
            return SupportedTrafficSignCountry.BELGIUM
        elif SupportedTrafficSignCountry.ZAMUNDA.value == country:
            return SupportedTrafficSignCountry.ZAMUNDA
        else:
            warnings.warn("Unknown country: Default traffic sign IDs are used. Specified country: " + country)
            return SupportedTrafficSignCountry.ZAMUNDA


class LaneletFactory:
    """ Class to create an object of class Lanelet from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Lanelet:
        """
        :param xml_node: XML element
        :return: object of class Lanelet according to the CommonRoad specification.
        """
        lanelet_id = int(xml_node.get('id'))

        left_vertices = cls._vertices(xml_node.find('leftBound'))
        right_vertices = cls._vertices(xml_node.find('rightBound'))
        center_vertices = 0.5 * (left_vertices + right_vertices)

        line_marking_left_vertices = cls._line_marking(xml_node.find('leftBound'))
        line_marking_right_vertices = cls._line_marking(xml_node.find('rightBound'))

        predecessors = cls._predecessors(xml_node)
        successors = cls._successors(xml_node)

        adjacent_left, adjacent_left_same_direction = cls._adjacent_left(xml_node)
        adjacent_right, adjacent_right_same_direction = cls._adjacent_right(xml_node)

        stop_line = cls._stop_line(xml_node, left_vertices[-1], right_vertices[-1])
        lanelet_type = cls._lanelet_type(xml_node)
        user_one_way = cls._user_one_way(xml_node)
        user_bidirectional = cls._user_bidirectional(xml_node)
        traffic_signs = cls._traffic_signs(xml_node)
        traffic_lights = cls._traffic_lights(xml_node)

        if cls._speed_limit_exists(xml_node) is not None:
            speed_limit = cls._speed_limit_exists(xml_node)
            if not hasattr(cls, '_speed_limits'):
                cls._speed_limits = {speed_limit: {lanelet_id}}
            elif cls._speed_limits.get(speed_limit) is not None:
                cls._speed_limits[speed_limit].add(lanelet_id)
            else:
                cls._speed_limits[speed_limit] = {lanelet_id}

        return Lanelet(
            left_vertices=left_vertices, center_vertices=center_vertices, right_vertices=right_vertices,
            lanelet_id=lanelet_id,
            predecessor=predecessors, successor=successors,
            adjacent_left=adjacent_left, adjacent_left_same_direction=adjacent_left_same_direction,
            adjacent_right=adjacent_right, adjacent_right_same_direction=adjacent_right_same_direction,
            line_marking_left_vertices=line_marking_left_vertices,
            line_marking_right_vertices=line_marking_right_vertices, stop_line=stop_line,
            lanelet_type=lanelet_type, user_one_way=user_one_way, user_bidirectional=user_bidirectional,
            traffic_signs=traffic_signs, traffic_lights=traffic_lights)

    @classmethod
    def _vertices(cls, xml_node: ElementTree.Element) -> np.ndarray:
        """
        Reads the vertices of the lanelet boundary.
        :param xml_node: XML element
        :return: The vertices of the boundary of the Lanelet described as a polyline
        """
        return PointListFactory.create_from_xml_node(xml_node)

    @classmethod
    def _predecessors(cls, xml_node: ElementTree.Element) -> List[int]:
        """
        Reads all predecessor lanelets.
        :param xml_node: XML element
        :return: list of IDs of all predecessor lanelets
        """
        predecessors = list()
        for l in xml_node.findall('predecessor'):
            predecessors.append(int(l.get('ref')))
        return predecessors

    @classmethod
    def _successors(cls, xml_node: ElementTree.Element) -> List[int]:
        """
        Reads all successor lanelets.
        :param xml_node: XML element
        :return: list of IDs of all successor lanelets
        """
        successors = list()
        for l in xml_node.findall('successor'):
            successors.append(int(l.get('ref')))
        return successors

    @classmethod
    def _adjacent_left(cls, xml_node: ElementTree.Element) -> Tuple[Union[int, None], Union[bool, None]]:
        """
        Reads the ID and the driving direction of a neighboring lanelet which is adjacent to the left.
        :param xml_node: XML element
        :return: the ID of the lanelet which is adjacent to the left (None if not existing);
                 the driving direction of the neighboring lanelet (None if not existing)
        """
        adjacent_left = None
        adjacent_left_same_direction = None
        if xml_node.find('adjacentLeft') is not None:
            adjacent_left = int(xml_node.find('adjacentLeft').get('ref'))
            if xml_node.find('adjacentLeft').get('drivingDir') == 'same':
                adjacent_left_same_direction = True
            else:
                adjacent_left_same_direction = False
        return adjacent_left, adjacent_left_same_direction

    @classmethod
    def _adjacent_right(cls, xml_node: ElementTree.Element) -> Tuple[Union[int, None], Union[bool, None]]:
        """
        Reads the ID and the driving direction of a neighboring lanelet which is adjacent to the right.
        :param xml_node: XML element
        :return: the ID of the lanelet which is adjacent to the right (None if not existing);
                 the driving direction of the neighboring lanelet (None if not existing)
        """
        adjacent_right = None
        adjacent_right_same_direction = None
        if xml_node.find('adjacentRight') is not None:
            adjacent_right = int(xml_node.find('adjacentRight').get('ref'))
            if xml_node.find('adjacentRight').get('drivingDir') == 'same':
                adjacent_right_same_direction = True
            else:
                adjacent_right_same_direction = False
        return adjacent_right, adjacent_right_same_direction

    @classmethod
    def _speed_limit_exists(cls, xml_node: ElementTree.Element) -> Union[float, None]:
        """
        Evaluates if a speed limit element from a previous CommonRoad version exists
        :param xml_node: XML element
        :return: boolean indicating if speed limit element exists
        """
        speed_limit = None
        if xml_node.find('speedLimit') is not None:
            speed_limit = float(xml_node.find('speedLimit').text)
            return speed_limit
        return speed_limit

    @classmethod
    def _lanelet_type(cls, xml_node: ElementTree.Element) -> Union[Set[LaneletType], None]:
        """
        Reads the lanelet types of the lanelet.

        :param xml_node: XML element
        :return: set of lanelet types for a lanelet
        """
        lanelet_types = set()
        for l_type in xml_node.findall('laneletType'):
            if LaneletType(l_type.text) is not None:
                lanelet_types.add(LaneletType(l_type.text))
            else:
                raise ValueError('<LaneletFactory/_lanelet_type>: Unkown type of lanelet: %s.' % l_type.text)
        return lanelet_types

    @classmethod
    def _user_one_way(cls, xml_node: ElementTree.Element) -> Union[Set[RoadUser], None]:
        """
        Reads the one way users of the lanelet.

        :param xml_node: XML element
        :return: set of allowed road users driving along the driving direction of a lanelet
        """
        users_one_way = set()
        for user in xml_node.findall('userOneWay'):
            if RoadUser(user.text) is not None:
                users_one_way.add(RoadUser(user.text))
            else:
                raise ValueError('<LaneletFactory/_user_one_way>: Unkown type of road user on lanelet: %s.' % user.text)

        return users_one_way

    @classmethod
    def _user_bidirectional(cls, xml_node: ElementTree.Element) -> Union[Set[RoadUser], None]:
        """
        Reads bidirectional users of the lanelet.

        :param xml_node: XML element
        :return: set of allowed road users driving in both driving directions of a lanelet
        """
        users_bidirectional = set()
        for user in xml_node.findall('userBidirectional'):
            if RoadUser(user.text) is not None:
                users_bidirectional.add(RoadUser(user.text))
            else:
                raise ValueError('<LaneletFactory/_user_bidirectional>: Unkown type of road user on lanelet: %s.'
                                 % user.text)
        return users_bidirectional

    @classmethod
    def _stop_line(cls, xml_node: ElementTree.Element, left_lanelet_end, right_lanelet_end) -> Union[StopLine, None]:
        """
        Reads the stop line of the lanelet.

        :param xml_node: XML element
        :return: stop line element of lanelet (None if not specified)
        """
        stop_line = None
        traffic_sign_ref = None
        traffic_light_ref = None
        if xml_node.find('stopLine') is not None:
            points = PointListFactory.create_from_xml_node(xml_node.find('stopLine'))
            line_marking = LineMarking(xml_node.find('stopLine').find('lineMarking').text)
            if xml_node.find('stopLine').find('trafficSignRef') is not None:
                traffic_sign_ref = set()
                for sign in xml_node.find('stopLine').findall('trafficSignRef'):
                    traffic_sign_ref.add(int(sign.get('ref')))
            if xml_node.find('stopLine').find('trafficLightRef') is not None:
                traffic_light_ref = set()
                for light in xml_node.find('stopLine').findall('trafficLightRef'):
                    traffic_light_ref.add(int(light.get('ref')))
            if len(points) > 0:
                stop_line = StopLine(start=points[0], end=points[1], line_marking=line_marking,
                                     traffic_sign_ref=traffic_sign_ref, traffic_light_ref=traffic_light_ref)
            else:
                stop_line = StopLine(start=left_lanelet_end, end=right_lanelet_end, line_marking=line_marking,
                                     traffic_sign_ref=traffic_sign_ref, traffic_light_ref=traffic_light_ref)

        return stop_line

    @classmethod
    def _line_marking(cls, xml_node: ElementTree.Element) -> Union[None, LineMarking]:
        """
        Reads the line marking of the left or right lanelet boundary.
        :param xml_node: XML element
        :return: the type of the line marking of the lanelet boundary (None if not specified).
        """
        line_marking = LineMarking.UNKNOWN
        if xml_node.find('lineMarking') is not None:
            if LineMarking(xml_node.find('lineMarking').text) is not None:
                line_marking = LineMarking(xml_node.find('lineMarking').text)
            else:
                raise ValueError('<LaneletFactory/_line_marking>: Unkown type of line marking: %s.'
                                 % line_marking)
        return line_marking

    @classmethod
    def _traffic_signs(cls, xml_node: ElementTree.Element) -> Union[Set[int], None]:
        """
        Reads the traffic sign references of the lanelet.

        :param xml_node: XML element
        :return: set of traffic sign IDs (None if not specified).
        """
        traffic_signs = set()
        for traffic_sign_ref in xml_node.findall('trafficSignRef'):
            if traffic_sign_ref.get("ref") is not None:
                traffic_signs.add(int(traffic_sign_ref.get("ref")))
            else:
                raise ValueError('<LaneletFactory/_traffic_signs>: Unknown type of traffic sign reference: %s.'
                                 % traffic_sign_ref.get("ref"))
        return traffic_signs

    @classmethod
    def _traffic_lights(cls, xml_node: ElementTree.Element) -> Union[Set[int], None]:
        """
        Reads the traffic sign references of the lanelet.

        :param xml_node: XML element
        :return: set of traffic light IDs (None if not specified).
        """
        traffic_lights = set()
        for traffic_light_ref in xml_node.findall('trafficLightRef'):
            if traffic_light_ref.get("ref") is not None:
                traffic_lights.add(int(traffic_light_ref.get("ref")))
            else:
                raise ValueError('<LaneletFactory/_traffic_signs>: Unkown type of traffic light reference: %s.'
                                 % traffic_light_ref.get("ref"))
        return traffic_lights


class TrafficSignFactory:
    """ Class to create an object of class TrafficSign from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, country: SupportedTrafficSignCountry,
                             first_traffic_sign_occurence: Dict[int, Set[int]],
                             lanelet_network: LaneletNetwork) -> TrafficSign:
        """
        :param xml_node: XML element
        :param country: country where traffic sign stands
        :param first_traffic_sign_occurence: set of first occurences of traffic sign
        :param lanelet_network: CommonRoad lanelet network
        :return: object of class TrafficSign according to the CommonRoad specification.
        """
        traffic_sign_id = int(xml_node.get('id'))
        assert traffic_sign_id in first_traffic_sign_occurence.keys(), \
            '<CommonRoadFileReader/TrafficSignFactory.create_from_xml_node>: ' \
            'CommonRoad file is invalid! Traffic sign {} is not referenced by a lanelet!'.format(traffic_sign_id)

        traffic_sign_elements = []
        for element in xml_node.findall('trafficSignElement'):
            traffic_sign_elements.append(TrafficSignElementFactory.create_from_xml_node(element, country))

        if xml_node.find('position') is not None:
            position = PointFactory.create_from_xml_node(xml_node.find('position').find('point'))
        else:
            # traffic signs are always placed on right side of road if no position is given (for right-hand traffic)
            position = None
            for lanelet_id in first_traffic_sign_occurence.get(traffic_sign_id):
                if country.value in LEFT_HAND_TRAFFIC:
                    if lanelet_network.find_lanelet_by_id(lanelet_id).adj_left_same_direction is None or \
                            lanelet_network.find_lanelet_by_id(lanelet_id).adj_left_same_direction is False:
                        if any(element.traffic_sign_element_id.name in TRAFFIC_SIGN_VALIDITY_START
                               for element in traffic_sign_elements):
                            position = lanelet_network.find_lanelet_by_id(lanelet_id).left_vertices[0]
                        else:
                            position = lanelet_network.find_lanelet_by_id(lanelet_id).left_vertices[-1]
                else:
                    if lanelet_network.find_lanelet_by_id(lanelet_id).adj_right_same_direction is None \
                            or lanelet_network.find_lanelet_by_id(lanelet_id).adj_right_same_direction is False:
                        if any(element.traffic_sign_element_id.name in TRAFFIC_SIGN_VALIDITY_START
                               for element in traffic_sign_elements):
                            position = lanelet_network.find_lanelet_by_id(lanelet_id).right_vertices[0]
                        else:
                            position = lanelet_network.find_lanelet_by_id(lanelet_id).right_vertices[-1]
            if position is None:
                current_lanelet =\
                    lanelet_network.find_lanelet_by_id(list(first_traffic_sign_occurence.get(traffic_sign_id))[0])
                if country.value in LEFT_HAND_TRAFFIC:
                    while current_lanelet.adj_left_same_direction is not None \
                            and current_lanelet.adj_left_same_direction is not False:
                        current_lanelet = lanelet_network.find_lanelet_by_id(current_lanelet.adj_left)
                    if any(element.traffic_sign_element_id.name in TRAFFIC_SIGN_VALIDITY_START for element in
                           traffic_sign_elements):
                        position = current_lanelet.left_vertices[0]
                    else:
                        position = current_lanelet.left_vertices[-1]
                else:
                    while current_lanelet.adj_right_same_direction is not None \
                            and current_lanelet.adj_right_same_direction is not False:
                        current_lanelet = lanelet_network.find_lanelet_by_id(current_lanelet.adj_right)
                    if any(element.traffic_sign_element_id.name in TRAFFIC_SIGN_VALIDITY_START for element in
                           traffic_sign_elements):
                        position = current_lanelet.right_vertices[0]
                    else:
                        position = current_lanelet.right_vertices[-1]

        if xml_node.get('virtual') is not None:
            if xml_node.get('virtual').text == "true":
                virtual = True
            elif xml_node.get('virtual').text == "false":
                virtual = False
            else:
                raise ValueError()
        else:
            virtual = False

        return TrafficSign(traffic_sign_id=traffic_sign_id, position=position,
                           first_occurrence=first_traffic_sign_occurence[traffic_sign_id],
                           traffic_sign_elements=traffic_sign_elements, virtual=virtual)


class TrafficSignElementFactory:
    """ Class to create an object of class TrafficSignElement from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element,
                             country: SupportedTrafficSignCountry) -> TrafficSignElement:
        """
        :param xml_node: XML element
        :param country: country where traffic sign stands
        :return: object of class TrafficSignElement according to the CommonRoad specification.
        """
        try:
            if country is SupportedTrafficSignCountry.GERMANY:
                traffic_sign_element_id = TrafficSignIDGermany(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.ZAMUNDA:
                traffic_sign_element_id = TrafficSignIDZamunda(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.USA:
                traffic_sign_element_id = TrafficSignIDUsa(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.CHINA:
                traffic_sign_element_id = TrafficSignIDChina(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.SPAIN:
                traffic_sign_element_id = TrafficSignIDSpain(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.RUSSIA:
                traffic_sign_element_id = TrafficSignIDRussia(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.ARGENTINA:
                traffic_sign_element_id = TrafficSignIDArgentina(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.ITALY:
                traffic_sign_element_id = TrafficSignIDItaly(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.FRANCE:
                traffic_sign_element_id = TrafficSignIDFrance(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.PUERTO_RICO:
                traffic_sign_element_id = TrafficSignIDPuertoRico(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.CROATIA:
                traffic_sign_element_id = TrafficSignIDCroatia(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.GREECE:
                traffic_sign_element_id = TrafficSignIDGreece(xml_node.find('trafficSignID').text)
            elif country is SupportedTrafficSignCountry.BELGIUM:
                traffic_sign_element_id = TrafficSignIDBelgium(xml_node.find('trafficSignID').text)
            else:
                warnings.warn("Unknown country: Default traffic sign ID is used. Specified country: " + country.value)
                traffic_sign_element_id = TrafficSignIDZamunda(xml_node.find('trafficSignID').text)
        except ValueError:
            warnings.warn("<FileReader>: Unknown TrafficElementID! Default traffic sign ID is used. Specified country: "
                          + country.value + " / Specified traffic sign ID: " + xml_node.find('trafficSignID').text)
            traffic_sign_element_id = TrafficSignIDZamunda.UNKNOWN

        additional_values = []
        for additional_value in xml_node.findall('additionalValue'):
            additional_values.append(additional_value.text)

        return TrafficSignElement(traffic_sign_element_id=traffic_sign_element_id,
                                  additional_values=additional_values)


class TrafficLightFactory:
    """ Class to create an object of class TrafficLight from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, country: SupportedTrafficSignCountry,
                             lanelet_network: LaneletNetwork) -> TrafficLight:
        """
        :param xml_node: XML element
        :param country: country where traffic sign stands
        :param lanelet_network: CommonRoad lanelet network
        :return: object of class TrafficLight according to the CommonRoad specification.
        """
        traffic_light_id = int(xml_node.get('id'))

        if xml_node.find('position') is not None:
            position = PointFactory.create_from_xml_node(xml_node.find('position').find('point'))
        else:
            # traffic lights are always placed on right side of road if no position is given (for right-hand traffic)
            current_lanelet = None
            for lanelet in lanelet_network.lanelets:
                if traffic_light_id in lanelet.traffic_lights:
                    current_lanelet = lanelet
                    break
            if current_lanelet is None:
                raise ValueError("Error in xml-file: Traffic Light not referenced by a lanelet")
            if country.value in LEFT_HAND_TRAFFIC:
                while current_lanelet.adj_left_same_direction is not None \
                        and current_lanelet.adj_left_same_direction is not False:
                    current_lanelet = lanelet_network.find_lanelet_by_id(current_lanelet.adj_left)
                position = current_lanelet.left_vertices[-1]
            else:
                while current_lanelet.adj_right_same_direction is not None \
                        and current_lanelet.adj_right_same_direction is not False:
                    current_lanelet = lanelet_network.find_lanelet_by_id(current_lanelet.adj_right)
                position = current_lanelet.right_vertices[-1]

        if xml_node.find('active') is not None:
            if xml_node.find('active').text == "true":
                active = True
            elif xml_node.find('active').text == "false":
                active = False
            else:
                active = True
        else:
            active = True

        if xml_node.find('direction') is not None:
            if xml_node.find('direction').text == "right":
                direction = TrafficLightDirection.RIGHT
            elif xml_node.find('direction').text == "straight":
                direction = TrafficLightDirection.STRAIGHT
            elif xml_node.find('direction').text == "left":
                direction = TrafficLightDirection.LEFT
            elif xml_node.find('direction').text == "leftStraight":
                direction = TrafficLightDirection.LEFT_STRAIGHT
            elif xml_node.find('direction').text == "straightRight":
                direction = TrafficLightDirection.STRAIGHT_RIGHT
            elif xml_node.find('direction').text == "leftRight":
                direction = TrafficLightDirection.LEFT_RIGHT
            elif xml_node.find('direction').text == "all":
                direction = TrafficLightDirection.ALL
            else:
                direction = TrafficLightDirection.ALL
        else:
            direction = TrafficLightDirection.ALL

        traffic_ligth_cycle, time_offset = TrafficLightCycleFactory.create_from_xml_node(xml_node.find('cycle'))

        return TrafficLight(traffic_light_id=traffic_light_id, cycle=traffic_ligth_cycle, position=position,
                            direction=direction, active=active, time_offset=time_offset)


class TrafficLightCycleFactory:
    """ Class to create an object of class TrafficLightCycleElement from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Tuple[List[TrafficLightCycleElement], int]:
        """
        :param xml_node: XML element
        :return: list of objects of class TrafficLightCycleElement according to the CommonRoad specification.
        """
        traffic_ligth_cycle_elements = []
        for cycleElement in xml_node.findall('cycleElement'):
            state = cycleElement.find('color').text
            duration = int(cycleElement.find('duration').text)
            traffic_ligth_cycle_elements.append(TrafficLightCycleElement(state=TrafficLightState(state),
                                                                         duration=duration))

        if xml_node.find('timeOffset') is not None:
            time_offset = int(xml_node.find('timeOffset').text)
        else:
            time_offset = 0

        return traffic_ligth_cycle_elements, time_offset


class IntersectionFactory:
    """ Class to create an object of class Intersection from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Intersection:
        """
        :param xml_node: XML element
        :return: object of class Intersection according to the CommonRoad specification.
        """
        intersection_id = int(xml_node.get('id'))
        incomings = []
        for incoming_node in xml_node.findall('incoming'):
            incomings.append(IntersectionIncomingFactory.create_from_xml_node(incoming_node))

        if xml_node.find('crossing') is not None:
            crossings = set()
            for crossing_ref in xml_node.find('crossing').findall('crossingLanelet'):
                crossings.add(int(crossing_ref.get("ref")))
        else:
            crossings = None

        return Intersection(intersection_id=intersection_id, incomings=incomings, crossings=crossings)


class IntersectionIncomingFactory:
    """ Class to create an object of class IntersectionIncomingElement from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> IntersectionIncomingElement:
        """
        :param xml_node: XML element
        :return: object of class IntersectionIncomingElement according to the CommonRoad specification.
        """
        incoming_id = int(xml_node.get('id'))
        incoming_lanelets = set()
        successors_right = set()
        successors_straight = set()
        successors_left = set()
        left_of = None
        for incoming_lanelet_ref in xml_node.findall('incomingLanelet'):
            incoming_lanelets.add(int(incoming_lanelet_ref.get('ref')))
        for successor_right_ref in xml_node.findall('successorsRight'):
            successors_right.add(int(successor_right_ref.get('ref')))
        for successor_straight_ref in xml_node.findall('successorsStraight'):
            successors_straight.add(int(successor_straight_ref.get('ref')))
        for successor_left_ref in xml_node.findall('successorsLeft'):
            successors_left.add(int(successor_left_ref.get('ref')))
        for left_of_ref in xml_node.findall('isLeftOf'):
            left_of = int(left_of_ref.get('ref'))

        return IntersectionIncomingElement(incoming_id=incoming_id, incoming_lanelets=incoming_lanelets,
                                           successors_right=successors_right, successors_straight=successors_straight,
                                           successors_left=successors_left, left_of=left_of)


class ObstacleFactory(ABC):
    @classmethod
    def read_type(cls, xml_node: ElementTree.Element) -> ObstacleType:
        obstacle_type = None
        if xml_node.find('type') is not None:
            if ObstacleType(xml_node.find('type').text) is not None:
                obstacle_type = ObstacleType(xml_node.find('type').text)
            else:
                raise ValueError('Type of obstacle is unknown. Got type: {}'.format(xml_node.find('type').text))

        return obstacle_type

    @classmethod
    def read_id(cls, xml_node: ElementTree.Element) -> int:
        obstacle_id = int(xml_node.get('id'))
        return obstacle_id

    @classmethod
    def read_initial_state(cls, xml_node: ElementTree.Element) -> State:
        initial_state = StateFactory.create_from_xml_node(xml_node)
        return initial_state

    @classmethod
    def read_shape(cls, xml_node: ElementTree.Element) -> Shape:
        shape = ShapeFactory.create_from_xml_node(xml_node)
        return shape

    @classmethod
    def read_initial_signal_state(cls, xml_node: ElementTree.Element) -> SignalState:
        initial_signal_state = SignalStateFactory.create_from_xml_node(xml_node)
        return initial_signal_state


class StaticObstacleFactory(ObstacleFactory):
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork,
                             lanelet_assignment: bool) -> StaticObstacle:
        obstacle_type = StaticObstacleFactory.read_type(xml_node)
        obstacle_id = StaticObstacleFactory.read_id(xml_node)
        initial_state = StaticObstacleFactory.read_initial_state(xml_node.find('initialState'))
        initial_signal_state = StaticObstacleFactory.read_initial_signal_state(xml_node.find('initialSignalState'))
        signal_series = SignalSeriesFactory.create_from_xml_node((xml_node.find('signalSeries')))
        shape = StaticObstacleFactory.read_shape(xml_node.find('shape'))

        if lanelet_assignment is True:
            rotated_shape = shape.rotate_translate_local(initial_state.position, initial_state.orientation)
            initial_shape_lanelet_ids = set(lanelet_network.find_lanelet_by_shape(rotated_shape))
            initial_center_lanelet_ids = set(lanelet_network.find_lanelet_by_position([initial_state.position])[0])
            for l_id in initial_shape_lanelet_ids:
                 lanelet_network.find_lanelet_by_id(l_id).add_static_obstacle_to_lanelet(obstacle_id=obstacle_id)
        else:
            initial_center_lanelet_ids = None
            initial_shape_lanelet_ids = None

        return StaticObstacle(obstacle_id=obstacle_id, obstacle_type=obstacle_type,
                              obstacle_shape=shape, initial_state=initial_state,
                              initial_center_lanelet_ids=initial_center_lanelet_ids,
                              initial_shape_lanelet_ids=initial_shape_lanelet_ids,
                              initial_signal_state=initial_signal_state,
                              signal_series=signal_series)


class DynamicObstacleFactory(ObstacleFactory):
    @staticmethod
    def find_obstacle_shape_lanelets(initial_state: State, state_list: List[State], lanelet_network: LaneletNetwork,
                                     obstacle_id: int, shape: Shape) -> Dict[int, Set[int]]:
        """
        Extracts for each shape the corresponding lanelets it is on

        :param initial_state: initial CommonRoad state
        :param state_list: trajectory state list
        :param lanelet_network: CommonRoad lanelet network
        :param obstacle_id: ID of obstacle
        :param shape: shape of obstacle
        :return: list of IDs of all predecessor lanelets
        """
        compl_state_list = [initial_state] + state_list
        lanelet_ids_per_state = {}

        for state in compl_state_list:
            rotated_shape = shape.rotate_translate_local(state.position, state.orientation)
            lanelet_ids = lanelet_network.find_lanelet_by_shape(rotated_shape)
            for l_id in lanelet_ids:
                lanelet_network.find_lanelet_by_id(l_id).add_dynamic_obstacle_to_lanelet(obstacle_id=obstacle_id,
                                                                                         time_step=state.time_step)
            lanelet_ids_per_state[state.time_step] = set(lanelet_ids)

        return lanelet_ids_per_state

    @staticmethod
    def find_obstacle_center_lanelets(initial_state: State, state_list: List[State],
                                      lanelet_network: LaneletNetwork) -> Dict[int, Set[int]]:
        """
        Extracts for each shape the corresponding lanelets it is on

        :param initial_state: initial CommonRoad state
        :param state_list: trajectory state list
        :param lanelet_network: CommonRoad lanelet network
        :return: list of IDs of all predecessor lanelets
        """
        compl_state_list = [initial_state] + state_list
        lanelet_ids_per_state = {}

        for state in compl_state_list:
            lanelet_ids = lanelet_network.find_lanelet_by_position([state.position])[0]
            lanelet_ids_per_state[state.time_step] = set(lanelet_ids)

        return lanelet_ids_per_state

    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork,
                             lanelet_assignment: bool) -> DynamicObstacle:
        obstacle_type = DynamicObstacleFactory.read_type(xml_node)
        obstacle_id = DynamicObstacleFactory.read_id(xml_node)
        shape = DynamicObstacleFactory.read_shape(xml_node.find('shape'))
        initial_state = DynamicObstacleFactory.read_initial_state(xml_node.find('initialState'))
        initial_signal_state = DynamicObstacleFactory.read_initial_signal_state(xml_node.find('initialSignalState'))
        signal_series = SignalSeriesFactory.create_from_xml_node((xml_node.find('signalSeries')))
        initial_center_lanelet_ids = set()
        initial_shape_lanelet_ids = set()

        if xml_node.find('trajectory') is not None:
            if lanelet_assignment is True:
                rotated_shape = shape.rotate_translate_local(initial_state.position, initial_state.orientation)
                initial_shape_lanelet_ids = set(lanelet_network.find_lanelet_by_shape(rotated_shape))
                initial_center_lanelet_ids = set(lanelet_network.find_lanelet_by_position([initial_state.position])[0])
                for l_id in initial_shape_lanelet_ids:
                    lanelet_network.find_lanelet_by_id(l_id).\
                        add_dynamic_obstacle_to_lanelet(obstacle_id=obstacle_id, time_step=initial_state.time_step)
            else:
                initial_shape_lanelet_ids = None
                initial_center_lanelet_ids = None
            trajectory = TrajectoryFactory.create_from_xml_node(xml_node.find('trajectory'))
            if lanelet_assignment is True:
                shape_lanelet_assignment = cls.find_obstacle_shape_lanelets(initial_state, trajectory.state_list,
                                                                            lanelet_network, obstacle_id, shape)
                center_lanelet_assignment = cls.find_obstacle_center_lanelets(initial_state, trajectory.state_list,
                                                                              lanelet_network)
            else:
                shape_lanelet_assignment = None
                center_lanelet_assignment = None
            prediction = TrajectoryPrediction(trajectory, shape, center_lanelet_assignment, shape_lanelet_assignment)
        elif xml_node.find('occupancySet') is not None:
            prediction = SetBasedPredictionFactory.create_from_xml_node(xml_node.find('occupancySet'))
        else:
            prediction = None
        return DynamicObstacle(obstacle_id=obstacle_id, obstacle_type=obstacle_type,
                               obstacle_shape=shape, initial_state=initial_state, prediction=prediction,
                               initial_center_lanelet_ids=initial_center_lanelet_ids,
                               initial_shape_lanelet_ids=initial_shape_lanelet_ids,
                               initial_signal_state=initial_signal_state,
                               signal_series=signal_series)


class TrajectoryFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) \
            -> Trajectory:
        state_list = list()
        for state_node in xml_node.findall('state'):
            state_list.append(StateFactory.create_from_xml_node(state_node))
        if isinstance(state_list[0].time_step, Interval):
            t0 = min(state_list[0].time_step)
        else:
            t0 = state_list[0].time_step
        return Trajectory(t0, state_list)


class SetBasedPredictionFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> SetBasedPrediction:
        occupancies = list()
        for occupancy in xml_node.findall('occupancy'):
            occupancies.append(OccupancyFactory.create_from_xml_node(occupancy))
        if isinstance(occupancies[0].time_step, Interval):
            t0 = min(occupancies[0].time_step)
        else:
            t0 = occupancies[0].time_step
        return SetBasedPrediction(t0, occupancies)


class OccupancyFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Occupancy:
        shape = ShapeFactory.create_from_xml_node(xml_node.find('shape'))
        time = read_time(xml_node.find('time'))
        return Occupancy(time, shape)


class ShapeFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Shape:
        shape_list = list()
        for c in list(xml_node):
            shape_list.append(cls._read_single_shape(c))
        shape = cls._create_shape_group_if_needed(shape_list)
        return shape

    @classmethod
    def _read_single_shape(cls, xml_node: ElementTree.Element) \
            -> Shape:
        tag_string = xml_node.tag
        if tag_string == 'rectangle':
            return RectangleFactory.create_from_xml_node(xml_node)
        elif tag_string == 'circle':
            return CircleFactory.create_from_xml_node(xml_node)
        elif tag_string == 'polygon':
            return PolygonFactory.create_from_xml_node(xml_node)

    @classmethod
    def _create_shape_group_if_needed(cls, shape_list: List[Shape]) -> Shape:
        if len(shape_list) > 1:
            sg = ShapeGroup(shape_list)
            return sg
        else:
            return shape_list[0]


class RectangleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Rectangle:
        length = float(xml_node.find('length').text)
        width = float(xml_node.find('width').text)
        if xml_node.find('orientation') is not None:
            orientation = float(xml_node.find('orientation').text)
        else:
            orientation = 0.0
        if xml_node.find('center') is not None:
            center = PointFactory.create_from_xml_node(
                xml_node.find('center'))
        else:
            center = np.array([0.0, 0.0])
        return Rectangle(length, width, center, orientation)


class CircleFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Circle:
        radius = float(xml_node.find('radius').text)
        if xml_node.find('center') is not None:
            center = PointFactory.create_from_xml_node(
                xml_node.find('center'))
        else:
            center = np.array([0.0, 0.0])
        return Circle(radius, center)


class PolygonFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Polygon:
        vertices = PointListFactory.create_from_xml_node(xml_node)
        return Polygon(vertices)


class PlanningProblemSetFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork) \
            -> PlanningProblemSet:
        planning_problem_set = PlanningProblemSet()
        for p in xml_node.findall('planningProblem'):
            planning_problem_set.add_planning_problem(
                PlanningProblemFactory.create_from_xml_node(p, lanelet_network))
        return planning_problem_set


class PlanningProblemFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork) \
            -> PlanningProblem:
        planning_problem_id = int(xml_node.get('id'))
        initial_state = cls._add_initial_state(xml_node)
        goal_region = GoalRegionFactory.create_from_xml_node(xml_node, lanelet_network)
        return PlanningProblem(planning_problem_id, initial_state, goal_region)

    @classmethod
    def _add_initial_state(cls, xml_node: ElementTree.Element) \
            -> State:
        initial_state = StateFactory.create_from_xml_node(xml_node.find('initialState'))
        return initial_state


class GoalRegionFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: LaneletNetwork)\
            -> GoalRegion:
        state_list = list()
        lanelets_of_goal_position = defaultdict(list)
        for idx, goal_state_node in enumerate(xml_node.findall('goalState')):
            state_list.append(StateFactory.create_from_xml_node(goal_state_node, lanelet_network))
            if goal_state_node.find('position') is not None\
                    and goal_state_node.find('position').find('lanelet') is not None:
                for l in goal_state_node.find('position').findall('lanelet'):
                    lanelets_of_goal_position[idx].append(int(l.get('ref')))
        if not lanelets_of_goal_position:
            lanelets_of_goal_position = None
        return GoalRegion(state_list, lanelets_of_goal_position)


class StateFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element, lanelet_network: Union[LaneletNetwork, None] = None)\
            -> State:
        state_args = dict()
        if xml_node.find('position') is not None:
            position = cls._read_position(xml_node.find('position'), lanelet_network)
            state_args['position'] = position
        if xml_node.find('time') is not None:
            state_args['time_step'] = read_time(xml_node.find('time'))
        if xml_node.find('orientation') is not None:
            orientation = cls._read_orientation(xml_node.find('orientation'))
            state_args['orientation'] = orientation
        if xml_node.find('velocity') is not None:
            speed = read_value_exact_or_interval(xml_node.find('velocity'))
            state_args['velocity'] = speed
        if xml_node.find('acceleration') is not None:
            acceleration = read_value_exact_or_interval(xml_node.find('acceleration'))
            state_args['acceleration'] = acceleration
        if xml_node.find('yawRate') is not None:
            yaw_rate = read_value_exact_or_interval(xml_node.find('yawRate'))
            state_args['yaw_rate'] = yaw_rate
        if xml_node.find('slipAngle') is not None:
            slip_angle = read_value_exact_or_interval(xml_node.find('slipAngle'))
            state_args['slip_angle'] = slip_angle
        return State(**state_args)

    @classmethod
    def _read_position(cls, xml_node: ElementTree.Element,
                       lanelet_network: Union[LaneletNetwork, None] = None) \
            -> Union[np.ndarray, Shape]:
        if xml_node.find('point') is not None:
            position = PointFactory.create_from_xml_node(xml_node.find('point'))
        elif (xml_node.find('rectangle') is not None
              or xml_node.find('circle') is not None
              or xml_node.find('polygon') is not None):
            position = ShapeFactory.create_from_xml_node(xml_node)
        elif lanelet_network is not None and xml_node.find('lanelet') is not None:
            position_list = list()
            for l in xml_node.findall('lanelet'):
                lanelet = lanelet_network.find_lanelet_by_id(int(l.get('ref')))
                polygon = lanelet.convert_to_polygon()
                position_list.append(polygon)
            position = ShapeGroup(position_list)
        else:
            raise Exception()
        return position

    @classmethod
    def _read_orientation(cls, xml_node: ElementTree.Element) -> Union[float, AngleInterval]:
        if xml_node.find('exact') is not None:
            value = float(xml_node.find('exact').text)
        elif xml_node.find('intervalStart') is not None \
                and xml_node.find('intervalEnd') is not None:
            value = AngleInterval(
                float(xml_node.find('intervalStart').text),
                float(xml_node.find('intervalEnd').text))
        else:
            raise Exception()
        return value


class SignalStateFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> Union[SignalState, None]:
        if xml_node is None:
            return None
        state_args = dict()
        if xml_node.find('time') is not None:
            state_args['time_step'] = read_time(xml_node.find('time'))
        if xml_node.find('horn') is not None:
            horn = cls._read_boolean(xml_node.find('horn'))
            state_args['horn'] = horn
        if xml_node.find('indicatorLeft') is not None:
            indicatorLeft = cls._read_boolean(xml_node.find('indicatorLeft'))
            state_args['indicator_left'] = indicatorLeft
        if xml_node.find('indicatorRight') is not None:
            indicatorRight = cls._read_boolean(xml_node.find('indicatorRight'))
            state_args['indicator_right'] = indicatorRight
        if xml_node.find('brakingLights') is not None:
            brakingLights = cls._read_boolean(xml_node.find('brakingLights'))
            state_args['braking_lights'] = brakingLights
        if xml_node.find('hazardWarningLights') is not None:
            hazardWarningLights = cls._read_boolean(xml_node.find('hazardWarningLights'))
            state_args['hazard_warning_lights'] = hazardWarningLights
        if xml_node.find('flashingBlueLights') is not None:
            flashingBlueLights = cls._read_boolean(xml_node.find('flashingBlueLights'))
            state_args['flashing_blue_lights'] = flashingBlueLights
        return SignalState(**state_args)

    @classmethod
    def _read_boolean(cls, xml_node: ElementTree.Element) -> bool:
        if xml_node.text == "true":
            return True
        elif xml_node.text == "false":
            return False
        else:
            raise ValueError()


class SignalSeriesFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> List[SignalState]:
        signal_state_list = []
        if xml_node is None:
            return signal_state_list
        for signal_state_node in xml_node.findall('signalState'):
            signal_state_list.append(SignalStateFactory.create_from_xml_node(signal_state_node))

        return signal_state_list


class PointListFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> np.ndarray:
        point_list = []
        for point_node in xml_node.findall("point"):
            point_list.append(
                PointFactory.create_from_xml_node(point_node))
        return np.array(point_list)


class PointFactory:
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> np.ndarray:
        x = float(xml_node.find('x').text)
        y = float(xml_node.find('y').text)
        if xml_node.find('z') is None:
            return np.array([x, y])
        else:
            z = float(xml_node.find('z').text)
            return np.array([x, y, z])


class EnvironmentObstacleFactory(ObstacleFactory):
    """ Class to create a list of objects of type EnvironmentObstacle from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> EnvironmentObstacle:
        obstacle_type = EnvironmentObstacleFactory.read_type(xml_node)
        obstacle_id = EnvironmentObstacleFactory.read_id(xml_node)
        shape = EnvironmentObstacleFactory.read_shape(xml_node.find('shape'))

        return EnvironmentObstacle(obstacle_id=obstacle_id, obstacle_type=obstacle_type, obstacle_shape=shape)


class PhantomObstacleFactory(ObstacleFactory):
    """ Class to create a list of objects of class PhantomObstacle from an XML element."""
    @classmethod
    def create_from_xml_node(cls, xml_node: ElementTree.Element) -> PhantomObstacle:
        obstacle_id = PhantomObstacleFactory.read_id(xml_node)
        if xml_node.find('occupancySet') is not None:
            prediction = SetBasedPredictionFactory.create_from_xml_node(xml_node.find('occupancySet'))
        else:
            prediction = None
        return PhantomObstacle(obstacle_id=obstacle_id, prediction=prediction)
